import mosaic.immutable_model
from mosaic.hdf5 import HDF5Store

M = mosaic.immutable_model
C = M.element('C')
atom = M.fragment('carbon', (), (('C', M.atom(C)),), ())
cube_universe = M.universe('cube', ((atom, 'carbons', 20),),
                           convention='my_own')

with HDF5Store("test.h5", "w") as store:
    store.store("universe", cube_universe)

with HDF5Store("test.h5") as store:
    r_universe = store.retrieve("universe")



# Package mosaic

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

from mosaic.version import version as __version__



# -*- coding: utf-8 -*-
"""Array model

.. moduleauthor:: Konrad Hinsen

An array model stores all data in NumPy arrays, in particular the tree
structure defining the fragments in universes. The array model is an
in-memory version of the HDF5 storage layout. It is used for HDF5 I/O,
but can also be of use for interfacing with C or Fortran code.

"""

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import operator

import numpy as N

import mosaic.api as api
from mosaic.utility import SymbolDict
from mosaic.utility import AbstractFactory
from mosaic.utility import MethodRegister
from mosaic.utility import isascii
from mosaic.utility import py_str

# dtypes for the various structured arrays
fragment_array_dtype = N.dtype([('parent_index', N.uint32),
                                ('label_symbol_index', N.uint32),
                                ('species_symbol_index', N.uint32),
                                # redundant:
                                ('number_of_fragments', N.uint32)])

atom_array_dtype = N.dtype([('parent_index', N.uint32),
                            ('label_symbol_index', N.uint32),
                            ('type_symbol_index', N.uint32),
                            ('name_symbol_index', N.uint32),
                            ('number_of_sites', N.uint32)])

bond_array_dtype = N.dtype([('atom_index_1', N.uint32),
                            ('atom_index_2', N.uint32),
                            ('bond_order_symbol_index', N.uint32)])

molecule_array_dtype = N.dtype([('fragment_index', N.uint32),
                                ('number_of_copies', N.uint32),
                                # redundant:
                                ('first_atom_index', N.uint32),
                                ('number_of_atoms', N.uint32),
                                ('first_bond_index', N.uint32),
                                ('number_of_bonds', N.uint32),
                                ('first_site_index', N.uint32),
                                ('number_of_sites', N.uint32)])

polymer_array_dtype = N.dtype([('fragment_index', N.uint32),
                               ('polymer_type_symbol_index', N.uint32)])

st_array_dtype = N.dtype([("rotation", N.float64, (3, 3)),
                          ("translation", N.float64, (3,))])


class Universe(api.MosaicUniverse):

    def __init__(self, universe):
        def count_fragments(f):
            return 1 + sum(count_fragments(ff) for ff in f.fragments)

        mol_fragments = tuple(fragment for fragment, _ in universe.molecules)
        natoms = sum(f.number_of_atoms for f in mol_fragments)
        nfragments = sum(count_fragments(f) for f in mol_fragments)
        nbonds = sum(f.number_of_bonds for f in mol_fragments)
        nmols = len(mol_fragments)

        symbol_id = SymbolDict()

        self._atom_array = N.empty((natoms,), atom_array_dtype)
        self._fragment_array = N.empty((nfragments + 1,),
                                       fragment_array_dtype)
        self._fragment_array[0] = (0, symbol_id[''], symbol_id[''], 1)
        self._bond_array = N.empty((nbonds,), bond_array_dtype)
        self._molecule_array = N.empty((nmols,), molecule_array_dtype)

        def store_fragment(f, iparent, ifrag, iatom, ibond, isite,
                           path, atom_index):
            self._fragment_array[ifrag]['parent_index'] = iparent
            self._fragment_array[ifrag]['label_symbol_index'] = \
                                                         symbol_id[f.label]
            self._fragment_array[ifrag]['species_symbol_index'] = \
                                                         symbol_id[f.species]
            iparent = ifrag
            ifrag += 1
            for ff in f.fragments:
                ifrag, iatom, ibond, isite = \
                       store_fragment(ff, iparent, ifrag, iatom, ibond, isite,
                                      '.'.join([path, ff.label]), atom_index)
            for a in f.atoms:
                atom_index['.'.join([path, a.label])] = iatom
                self._atom_array[iatom]['parent_index'] = iparent
                self._atom_array[iatom]['label_symbol_index'] = \
                                                symbol_id[a.label]
                self._atom_array[iatom]['type_symbol_index'] = \
                                                symbol_id[a.type]
                self._atom_array[iatom]['name_symbol_index'] = \
                                                symbol_id[a.name]
                self._atom_array[iatom]['number_of_sites'] = a.number_of_sites
                iatom += 1
                isite += a.number_of_sites
            bonds = []
            for a1, a2, order in f.bonds:
                i1 = atom_index['.'.join([path, a1])]
                i2 = atom_index['.'.join([path, a2])]
                if i1 > i2:
                    i1, i2 = i2, i1
                bonds.append((i1, i2, symbol_id[order]))
            bonds.sort()
            self._bond_array[ibond:ibond + len(bonds)] = bonds
            ibond += len(bonds)

            # ifrag now points to the first fragment that is not
            # a child of the current one. ifrag-iparent is thus
            # the number of fragments that are part of the current one.
            self._fragment_array[iparent]['number_of_fragments'] = \
                                                          ifrag - iparent
            return ifrag, iatom, ibond, isite

        ifrag = 1
        iatom = 0
        ibond = 0
        isite = 0
        imol = 0
        polymers = []
        for fragment, count in universe.molecules:
            self._molecule_array[imol]['fragment_index'] = ifrag
            self._molecule_array[imol]['number_of_copies'] = count
            self._molecule_array[imol]['first_atom_index'] = iatom
            self._molecule_array[imol]['first_bond_index'] = ibond
            self._molecule_array[imol]['first_site_index'] = isite
            if fragment.is_polymer:
                polymers.append((ifrag, symbol_id[fragment.polymer_type]))
            ifrag, iatom, ibond, isite = \
                   store_fragment(fragment, 0, ifrag, iatom, ibond, isite,
                                  '', {})
            self._molecule_array[imol]['number_of_atoms'] = \
                       iatom - self._molecule_array[imol]['first_atom_index']
            self._molecule_array[imol]['number_of_bonds'] = \
                       ibond - self._molecule_array[imol]['first_bond_index']
            self._molecule_array[imol]['number_of_sites'] = \
                       isite - self._molecule_array[imol]['first_site_index']
            imol += 1
        if len(polymers) > 0:
            self._polymer_array = N.array(polymers, dtype=polymer_array_dtype)
        else:
            self._polymer_array = N.zeros((0,), dtype=polymer_array_dtype)
        assert ifrag == nfragments + 1
        assert iatom == natoms
        assert ibond == nbonds

        # Other universe attributes
        self._cell_shape = universe.cell_shape
        self._convention = universe.convention
        # Numpy raises an exception if the transformation list is a
        # Python tuple AND it happens to be empty, so we convert
        # it to a list first.
        self._symmetry_transformations = \
                N.array(list(universe.symmetry_transformations),
                        dtype=st_array_dtype)
        symbols = list(symbol_id.items())
        symbols.sort(key=operator.itemgetter(1))
        self._symbols = tuple(s for s, i in symbols)

    # Create from existing arrays, typically read from HDF5
    @classmethod
    def from_arrays(cls, atom_array, bond_array, fragment_array,
                    polymer_array, molecule_array, symbols,
                    symmetry_transformations, cell_shape, convention):
        assert atom_array.dtype == atom_array_dtype
        assert bond_array.dtype == bond_array_dtype
        assert fragment_array.dtype == fragment_array_dtype
        if polymer_array is None:
            polymer_array = N.zeros((0,), dtype=polymer_array_dtype)
        assert polymer_array.dtype == polymer_array_dtype
        assert molecule_array.dtype == molecule_array_dtype
        assert isinstance(symbols, tuple)
        assert symmetry_transformations.dtype == st_array_dtype
        assert isascii(cell_shape)
        assert isascii(convention)

        universe = cls.__new__(cls)
        universe._atom_array = atom_array
        universe._bond_array = bond_array
        universe._fragment_array = fragment_array
        universe._polymer_array = polymer_array
        universe._molecule_array = molecule_array
        universe._symbols = tuple(symbols)
        universe._symmetry_transformations = symmetry_transformations
        universe._cell_shape = cell_shape
        universe._convention = convention
        return universe

    # API implementation
    @property
    def cell_shape(self):
        return self._cell_shape

    @property
    def symmetry_transformations(self):
        return tuple(tuple(t) for t in self._symmetry_transformations)

    @property
    def convention(self):
        return self._convention

    @property
    def molecules(self):
        return tuple((Fragment(self, m['fragment_index']),
                      int(m['number_of_copies']))
                     for m in self._molecule_array)

    # More efficient implementations of the number properties

    @property
    def number_of_molecules(self):
        return int(N.sum(self._molecule_array['number_of_copies']))

    @property
    def number_of_atoms(self):
        return int(N.sum(self._molecule_array['number_of_copies']
                         * self._molecule_array['number_of_atoms']))

    @property
    def number_of_sites(self):
        return int(N.sum(self._molecule_array['number_of_copies']
                         * self._molecule_array['number_of_sites']))

    @property
    def number_of_bonds(self):
        return int(N.sum(self._molecule_array['number_of_copies']
                         * self._molecule_array['number_of_bonds']))

    @property
    def number_of_template_atoms(self):
        return int(N.sum(self._molecule_array['number_of_atoms']))

    @property
    def number_of_template_sites(self):
        return int(N.sum(self._molecule_array['number_of_sites']))

    # More efficient implementation of various methods

    def recursive_atom_iterator(self):
        for mol in self._molecule_array:
            first = mol['first_atom_index']
            natoms = mol['number_of_atoms']
            for _ in range(mol['number_of_copies']):
                for iatom in range(first, first+natoms):
                    yield Atom(self, iatom)

    def bond_index_array(self):
        atoms = 0
        fatoms = 0
        bonds = []
        for mol in self._molecule_array:
            first = mol['first_bond_index']
            nbonds = mol['number_of_bonds']
            natoms = mol['number_of_atoms']
            a1 = self._bond_array[first:first+nbonds]['atom_index_1']
            a2 = self._bond_array[first:first+nbonds]['atom_index_2']
            fbonds = N.concatenate([a1[:, N.newaxis], a2[:, N.newaxis]],
                                    axis=1)
            for _ in range(mol['number_of_copies']):
                bonds.append(fbonds - fatoms + atoms)
                atoms += natoms
            fatoms += natoms
        return N.concatenate(bonds, axis=0)

class Fragment(api.MosaicFragment):

    def __init__(self, universe, ifrag):
        self._universe = universe
        self._ifrag = ifrag

    def __repr__(self):
        return "Fragment('%s', '%s')" % (self.label, self.species)

    @property
    def parent(self):
        ifrag = self._universe._fragment_array[self._ifrag]['parent_index']
        if ifrag == 0:
            return None
        else:
            return Fragment(self._universe, ifrag)

    # API implementation
    @property
    def label(self):
        isym = \
            self._universe._fragment_array[self._ifrag]['label_symbol_index']
        return self._universe._symbols[isym]

    @property
    def species(self):
        isym = \
            self._universe._fragment_array[self._ifrag]['species_symbol_index']
        return self._universe._symbols[isym]

    @property
    def fragments(self):
        mask = self._universe._fragment_array['parent_index'] == self._ifrag
        return [Fragment(self._universe, ifrag)
                for ifrag in N.repeat(N.arange(len(mask)), mask)]

    @property
    def atoms(self):
        mask = self._universe._atom_array['parent_index'] == self._ifrag
        return [Atom(self._universe, iatom)
                for iatom in N.repeat(N.arange(len(mask)), mask)]

    @property
    def bonds(self):

        universe = self._universe

        def ancestors(ip):
            anc = []
            while True:
                if ip == 0:
                    return None
                if ip == self._ifrag:
                    break
                anc.insert(0, ip)
                ip = universe._fragment_array[ip]['parent_index']
            return anc

        bonds = []
        for ia1, ia2, iorder in universe._bond_array[
                                ['atom_index_1', 'atom_index_2',
                                 'bond_order_symbol_index']]:
            anc1 = ancestors(universe._atom_array[ia1]['parent_index'])
            if anc1 is None:
                continue
            anc2 = ancestors(universe._atom_array[ia2]['parent_index'])
            if anc2 is None:
                continue
            if len(anc1) == 0 or len(anc2) == 0 or anc1[0] != anc2[0]:
                # bond belongs to the current fragment
                p1 = [universe._fragment_array[if1]['label_symbol_index']
                      for if1 in anc1]
                p1.append(universe._atom_array[ia1]['label_symbol_index'])
                p2 = [universe._fragment_array[if2]['label_symbol_index']
                      for if2 in anc2]
                p2.append(universe._atom_array[ia2]['label_symbol_index'])
                bonds.append(('.'.join(universe._symbols[l] for l in p1),
                              '.'.join(universe._symbols[l] for l in p2),
                              universe._symbols[iorder]))

        return tuple(bonds)

    @property
    def is_polymer(self):
        return N.sum(self._universe._polymer_array['fragment_index']
                     == self._ifrag) == 1

    @property
    def polymer_type(self):
        mask = self._universe._polymer_array['fragment_index'] == self._ifrag
        if mask.any():
            index = self._universe._polymer_array['polymer_type_symbol_index']
            isym = N.repeat(index, mask)[0]
            return self._universe._symbols[isym]
        else:
            raise ValueError("polymer_type not defined in a non-polymer")

    # More efficient implementation of the number properties

    @property
    def number_of_atoms(self):
        ifrag = self._ifrag
        nfrag = ifrag + \
                self._universe._fragment_array[ifrag]['number_of_fragments']
        return N.sum(N.logical_and(
                    self._universe._atom_array['parent_index'] >= ifrag,
                    self._universe._atom_array['parent_index'] < nfrag))

    @property
    def number_of_sites(self):
        ifrag = self._ifrag
        nfrag = ifrag + \
                self._universe._fragment_array[ifrag]['number_of_fragments']
        mask = N.logical_and(
                    self._universe._atom_array['parent_index'] >= ifrag,
                    self._universe._atom_array['parent_index'] < nfrag)
        return N.sum(N.repeat(self._universe._atom_array['number_of_sites'],
                              mask))

    @property
    def number_of_bonds(self):
        ifrag = self._ifrag
        nfrag = ifrag + \
                self._universe._fragment_array[ifrag]['number_of_fragments']
        a1 = N.take(self._universe._atom_array['parent_index'],
                    self._universe._bond_array['atom_index_1'])
        a2 = N.take(self._universe._atom_array['parent_index'],
                    self._universe._bond_array['atom_index_2'])
        return N.sum(N.logical_and(N.logical_and(a1 >= ifrag, a1 < nfrag),
                                   N.logical_and(a2 >= ifrag, a2 < nfrag)))


class Atom(api.MosaicAtom):

    def __init__(self, universe, iatom):
        self._universe = universe
        self._iatom = iatom

    def __repr__(self):
        return "Atom('%s')" % self.label

    @property
    def parent(self):
        ifrag = self._universe._atom_array[self._iatom]['parent_index']
        assert ifrag > 0
        return Fragment(self._universe, ifrag)

    # API implementation
    @property
    def label(self):
        isym = self._universe._atom_array[self._iatom]['label_symbol_index']
        return self._universe._symbols[isym]

    @property
    def type(self):
        isym_type = \
            self._universe._atom_array[self._iatom]['type_symbol_index']
        return self._universe._symbols[isym_type]

    @property
    def name(self):
        isym_name = \
            self._universe._atom_array[self._iatom]['name_symbol_index']
        return self._universe._symbols[isym_name]

    @property
    def number_of_atoms(self):
        return 1

    @property
    def number_of_sites(self):
        return int(self._universe._atom_array[self._iatom]['number_of_sites'])


class Property(api.MosaicProperty):

    def __init__(self, universe, name, units, data, property_type):
        api.validate_type(universe, Universe, "universe")
        api.validate_label(name, "name")
        api.validate_units(units, "units")
        api.validate_array(data, None, self._allowed_dtypes, "data")
        api.validate_value(property_type, self._allowed_types, "property_type")
        self._universe = universe
        self._name = name
        self._units = units
        self._data = data
        self._type = property_type

    # Data Model API implementation
    @property
    def universe(self):
        return self._universe

    @property
    def name(self):
        return self._name

    @property
    def units(self):
        return self._units

    @property
    def data(self):
        return self._data

    @property
    def type(self):
        return self._type


class Label(api.MosaicLabel):

    def __init__(self, universe, name, strings, label_type):
        api.validate_type(universe, Universe, "universe")
        api.validate_label(name, "name")
        api.validate_sequence(strings, str, "strings")
        api.validate_value(label_type, self._allowed_types, "label_type")
        self._universe = universe
        self._name = name
        self._string_array = N.array(list('\0'.join(strings) + '\0'),
                                     dtype='S')
        self._type = label_type

    @classmethod
    def from_arrays(cls, universe, name, array, label_type):
        label = cls.__new__(cls)
        label._universe = universe
        label._name = name
        label._string_array = array
        label._type = label_type
        return label

    # Data Model API implementation
    @property
    def universe(self):
        return self._universe

    @property
    def name(self):
        return self._name

    @property
    def strings(self):
        eos = N.repeat(N.arange(self._string_array.shape[0]),
                       self._string_array.view(dtype=N.uint8) == 0)
        strings = []
        f = 0
        for l in eos:
            strings.append(py_str(self._string_array[f:l].tostring()))
            f = l + 1
        return strings

    @property
    def type(self):
        return self._type


class Configuration(api.MosaicConfiguration):

    def __init__(self, universe, positions, cell_parameters):
        api.validate_type(universe, Universe, "universe")
        api.validate_array(positions, (universe.number_of_sites, 3),
                           self._allowed_dtypes, "positions")
        api.validate_array(cell_parameters,
                           universe._cell_parameter_array_shapes
                                                      [universe.cell_shape],
                           [positions.dtype], "cell_parameters")
        self._universe = universe
        self._positions = positions
        self._cell_parameters = cell_parameters

    # Data Model API implementation
    @property
    def universe(self):
        return self._universe

    @property
    def positions(self):
        return self._positions

    @property
    def cell_parameters(self):
        return self._cell_parameters


class Selection(api.MosaicSelection):

    def __init__(self, universe, indices, selection_type):
        api.validate_type(universe, Universe, "universe")
        api.validate_array(indices, None,
                           [N.uint8, N.uint16, N.uint32, N.uint64], "indices")
        api.validate_value(selection_type, self._allowed_types,
                           "selection_type")
        self._universe = universe
        self._indices = indices
        self._type = selection_type

    # Data Model API implementation
    @property
    def universe(self):
        return self._universe

    @property
    def indices(self):
        return self._indices

    @property
    def type(self):
        return self._type


class Factory(AbstractFactory):

    handler = MethodRegister()

    @handler(api.MosaicUniverse)
    def _make_universe(self, universe):
        return Universe(universe)

    @handler(api.MosaicProperty)
    def _make_property(self, property):
        universe = self(property.universe)
        return Property(universe, property.name, property.units,
                        property.data, property.type)

    @handler(api.MosaicLabel)
    def _make_label(self, label):
        universe = self(label.universe)
        return Label(universe, label.name, label.strings, label.type)

    @handler(api.MosaicConfiguration)
    def _make_configuration(self, configuration):
        universe = self(configuration.universe)
        return Configuration(universe,
                             configuration.positions,
                             configuration.cell_parameters)

    @handler(api.MosaicSelection)
    def _make_selection(self, selection):
        universe = self(selection.universe)
        return Selection(universe, selection.indices, selection.type)



# -*- coding: utf-8 -*-

# Tests for module mosaic.hdf5

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import copy
import os
import tempfile
import unittest

import numpy as N
import numpy.random as NR
import immutable.np as IN
import h5py

import mosaic.mutable_model
import mosaic.immutable_model
import mosaic.array_model
from mosaic.api import is_valid
from mosaic.hdf5 import HDF5Store


class TestHDF5Store(object):

    @classmethod
    def initialize(cls):
        cls.tempdir = tempfile.mkdtemp()
        cls.filename = os.path.join(cls.tempdir, 'test.h5')

    @classmethod
    def cleanup(cls):
        os.remove(cls.filename)
        os.removedirs(cls.tempdir)

    def __init__(self, mode='r'):
        self.file = h5py.File(self.filename, mode)
        self.store = HDF5Store(self.file)

    def close(self):
        del self.store
        self.file.close()

    def __enter__(self):
        return self.store

    def __exit__(self, type, value, traceback):
        self.close()
        return False


class PeptideTest(object):

    def test_universe(self):
        f = mosaic.array_model.Factory()
        for universe in [self.cube_universe, self.infinite_universe]:
            array_universe = f(universe)
            self.assertTrue(is_valid(array_universe))
            with TestHDF5Store('w') as store:
                store.store("universe", array_universe)
                self.assertTrue(store.retrieve("universe") is array_universe)
            with TestHDF5Store('r') as store:
                r_universe = store.retrieve("universe")
            self.assertFalse(r_universe is array_universe)
            self.assertFalse(r_universe is universe)
            self.assertTrue(r_universe.is_equivalent(array_universe))
            self.assertTrue(r_universe.is_equivalent(universe))
            self.assertTrue(is_valid(r_universe))

    def test_configuration(self):
        M = self.M

        conf = M.Configuration(self.cube_universe,
                               IN.array(NR.normal(size=(40,3))),
                               IN.array(10., N.float64))
        self.assertTrue(is_valid(conf))
        with TestHDF5Store('w') as store:
            self.assertRaises(IOError,
                              lambda : store.store("configuration", conf))
            store.store("universe", self.cube_universe)
            store.store("configuration", conf)
            self.assertTrue(store.retrieve("configuration") is conf)
        with TestHDF5Store('r') as store:
            conf_r = store.retrieve("configuration")
        self.assertFalse(conf_r is conf)
        self.assertTrue(conf_r.is_equivalent(conf))
        self.assertTrue(is_valid(conf_r))

        conf = M.Configuration(self.infinite_universe,
                               IN.array(NR.normal(size=(40,3))),
                               None)
        self.assertTrue(is_valid(conf))
        with TestHDF5Store('w') as store:
            self.assertRaises(IOError,
                              lambda : store.store("configuration", conf))
            store.store("universe", self.infinite_universe)
            store.store("configuration", conf)
            self.assertTrue(store.retrieve("configuration") is conf)
        with TestHDF5Store('r') as store:
            conf_r = store.retrieve("configuration")
        self.assertFalse(conf_r is conf)
        self.assertTrue(conf_r.is_equivalent(conf))
        self.assertTrue(is_valid(conf_r))

    def test_property(self):
        M = self.M
        for universe in[self.cube_universe, self.infinite_universe]:
            for p in [M.AtomProperty(universe, "mass", "amu",
                         IN.array(NR.normal(size=(universe.number_of_atoms,)))),
                      M.SiteProperty(universe, "foo", "",
                         IN.array(NR.normal(size=(universe.number_of_sites, 3)))),
                      M.TemplateAtomProperty(universe, "mass", "amu",
                         IN.array(NR.normal(size=(universe.number_of_template_atoms,
                                                  1)))),
                      M.TemplateSiteProperty(universe, "foo", "bar",
                         IN.array(NR.normal(size=(universe.number_of_template_sites,
                                                  2, 2)))),
                      ]:
                self.assertTrue(is_valid(p))
                with TestHDF5Store('w') as store:
                    self.assertRaises(IOError,
                                      lambda: store.store("property", p))
                    store.store("universe", universe)
                    store.store("property", p)
                    self.assertTrue(store.retrieve("property") is p)
                with TestHDF5Store('r') as store:
                    p_r = store.retrieve("property")
                self.assertFalse(p_r is p)
                self.assertTrue(p_r.is_equivalent(p))
                self.assertTrue(is_valid(p_r))

    def test_label(self):
        M = self.M
        for universe in[self.cube_universe, self.infinite_universe]:
            for l in [M.AtomLabel(universe, "element",
                                  tuple(a.name
                                        for f, n in universe.molecules
                                        for _ in range(n)
                                        for a in f.recursive_atom_iterator())),
                      M.TemplateAtomLabel(universe, "element",
                                          tuple(a.name
                                                for f, n in universe.molecules
                                                for a in f.recursive_atom_iterator())),
                      M.SiteLabel(universe, "element",
                                  tuple(a.name
                                        for f, n in universe.molecules
                                        for _ in range(n)
                                        for a in f.recursive_atom_iterator()
                                        for __ in range(a.number_of_sites))),
                      M.TemplateSiteLabel(universe, "element",
                                          tuple(a.name
                                                for f, n in universe.molecules
                                                for a in f.recursive_atom_iterator()
                                                for _ in range(a.number_of_sites))),
                      ]:
                self.assertTrue(is_valid(l))
                with TestHDF5Store('w') as store:
                    self.assertRaises(IOError,
                                      lambda: store.store("label", l))
                    store.store("universe", universe)
                    store.store("label", l)
                    self.assertTrue(store.retrieve("label") is l)
                with TestHDF5Store('r') as store:
                    l_r = store.retrieve("label")
                self.assertFalse(l_r is l)
                self.assertTrue(l_r.is_equivalent(l))
                self.assertTrue(is_valid(l_r))

    def test_selection(self):
        M = self.M
        universe = self.infinite_universe
        for s in [M.AtomSelection(universe, [0]),
                  M.SiteSelection(universe, [1, 2]),
                  M.TemplateAtomSelection(universe, [0, 3]),
                  M.TemplateSiteSelection(universe, [2, 3])]:
            self.assertTrue(is_valid(s))
            with TestHDF5Store('w') as store:
                self.assertRaises(IOError,
                                  lambda: store.store("selection", s))
                store.store("universe", universe)
                store.store("selection", s)
                self.assertTrue(store.retrieve("selection") is s)
            with TestHDF5Store('r') as store:
                s_r = store.retrieve("selection")
            self.assertFalse(s_r is s)
            self.assertTrue(s_r.is_equivalent(s))
            self.assertTrue(is_valid(s_r))


class MMPeptideTest(unittest.TestCase,
                    PeptideTest):
    M = mosaic.mutable_model

    def setUp(self):
        TestHDF5Store.initialize()

        M = self.M
        C = M.Element('C')
        H = M.Element('H')
        N = M.Element('N')
        O = M.Element('O')
        peptide_group = M.Fragment('peptide', 'peptide_group',
                                   (),
                                   (M.Atom('CA', C),
                                    M.Atom('HA', H),
                                    M.Atom('H', H),
                                    M.Atom('N', N),
                                    M.Atom('C', C),
                                    M.Atom('O', O)),
                                    (('N', 'H', 'single'),
                                     ('N', 'CA', 'single'),
                                     ('CA', 'HA', 'single'),
                                     ('CA', 'C', 'single'),
                                     ('C', 'O', 'double')))
        ala_sidechain = M.Fragment('sidechain', 'ala_sidechain',
                                   (),
                                   (M.Atom('CB', C),
                                    M.Atom('HB1', H),
                                    M.Atom('HB2', H),
                                    M.Atom('HB3', H)),
                                   (('CB', 'HB1', 'single'),
                                    ('CB', 'HB2', 'single'),
                                    ('CB', 'HB3', 'single'),))
        ala = lambda label: M.Fragment(label, 'alanine',
                                       (copy.copy(peptide_group),
                                        copy.copy(ala_sidechain)),
                                       (),
                                       (('peptide.CA',
                                         'sidechain.CB',
                                         'single'),))
        ala_dipeptide = M.Polymer('di_ala', 'alanine_dipeptide',
                                  (ala('ALA1'), ala('ALA2')),
                                  (('ALA1.peptide.C',
                                    'ALA2.peptide.N',
                                    'single'),),
                                  'polypeptide')
        self.cube_universe = M.Universe('cube', [(ala_dipeptide, 2)],
                                        convention='my_own')
        self.assertTrue(is_valid(self.cube_universe))
        self.infinite_universe = M.Universe('infinite', [(ala_dipeptide, 2)],
                                            convention='my_own')
        self.assertTrue(is_valid(self.infinite_universe))

    def tearDown(self):
        TestHDF5Store.cleanup()


class IMPeptideTest(unittest.TestCase,
                    PeptideTest):
    M = mosaic.immutable_model

    def setUp(self):
        TestHDF5Store.initialize()

        M = self.M
        C = M.element('C')
        H = M.element('H')
        N = M.element('N')
        O = M.element('O')
        peptide_group = M.fragment('peptide',
                                   (),
                                   (('CA', M.atom(C)),
                                    ('HA', M.atom(H)),
                                    ('H', M.atom(H)),
                                    ('N', M.atom(N)),
                                    ('C', M.atom(C)),
                                    ('O', M.atom(O))),
                                   (('N', 'H', "single"),
                                    ('N', 'CA', "single"),
                                    ('CA', 'HA', "single"),
                                    ('CA', 'C', "single"),
                                    ('C', 'O', "double")))
        ala_sidechain = M.fragment('ala_sidechain',
                                   (),
                                   (('CB', M.atom(C)),
                                    ('HB1', M.atom(H)),
                                    ('HB2', M.atom(H)),
                                    ('HB3', M.atom(H))),
                                   (('CB', 'HB1', "single"),
                                    ('CB', 'HB2', "single"),
                                    ('CB', 'HB3', "single"),))
        ala = M.fragment('alanine',
                         (('peptide', peptide_group),
                          ('sidechain', ala_sidechain)),
                         (),
                         (('peptide.CA', 'sidechain.CB', "single"),))
        ala_dipeptide = M.polymer('alanine_dipeptide',
                                  (('ALA1', ala),
                                   ('ALA2', ala)),
                                  (('ALA1.peptide.C', 'ALA2.peptide.N', "single"),),
                                  'polypeptide')
        self.cube_universe = M.universe('cube', ((ala_dipeptide, 'di_ala', 2),),
                                        convention='my_own')
        self.assertTrue(is_valid(self.cube_universe))
        self.infinite_universe = M.universe('infinite',
                                            ((ala_dipeptide, 'di_ala', 2),),
                                            convention='my_own')
        self.assertTrue(is_valid(self.infinite_universe))

    def tearDown(self):
        TestHDF5Store.cleanup()


class AtomicSystemTest(unittest.TestCase):
    M = mosaic.immutable_model

    def setUp(self):
        TestHDF5Store.initialize()

        M = self.M
        C = M.element('C')
        atom = M.fragment('carbon', (), (('C', M.atom(C)),), ())
        self.cube_universe = M.universe('cube', ((atom, 'carbons', 20),),
                                        convention='my_own')
        self.assertTrue(is_valid(self.cube_universe))

    def test_universe(self):
        f = mosaic.array_model.Factory()
        for universe in [self.cube_universe]:
            array_universe = f(universe)
            self.assertTrue(is_valid(array_universe))
            with TestHDF5Store('w') as store:
                store.store("universe", array_universe)
                self.assertTrue(store.retrieve("universe") is array_universe)
            with TestHDF5Store('r') as store:
                r_universe = store.retrieve("universe")
            self.assertFalse(r_universe is array_universe)
            self.assertFalse(r_universe is universe)
            self.assertTrue(r_universe.is_equivalent(array_universe))
            self.assertTrue(r_universe.is_equivalent(universe))
            self.assertTrue(is_valid(r_universe))

    def tearDown(self):
        TestHDF5Store.cleanup()


def suite():
    loader = unittest.TestLoader()
    s = unittest.TestSuite()
    s.addTest(loader.loadTestsFromTestCase(MMPeptideTest))
    s.addTest(loader.loadTestsFromTestCase(IMPeptideTest))
    s.addTest(loader.loadTestsFromTestCase(AtomicSystemTest))
    return s

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# Run the complete test suite

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import unittest

import basic_tests
import mutable_model_tests
import immutable_model_tests
import equivalence_tests
import hdf5_tests
import xml_tests
import pdb_tests

def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTests(basic_tests.suite())
    test_suite.addTests(mutable_model_tests.suite())
    test_suite.addTests(immutable_model_tests.suite())
    test_suite.addTests(equivalence_tests.suite())
    test_suite.addTests(hdf5_tests.suite())
    test_suite.addTests(xml_tests.suite())
    test_suite.addTests(pdb_tests.suite())
    return test_suite

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite())




# -*- coding: utf-8 -*-

# Equivalence tests between different models

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import unittest
import copy
import itertools as IT

import numpy as N
import immutable.np as IN

import mosaic.immutable_model as IM
import mosaic.mutable_model as MM
import mosaic.array_model as AM
from mosaic.api import is_valid

class WaterTest(unittest.TestCase):

    def make_mutable_universe(self):
        water = MM.Fragment("solvent", "water", (),
                            (MM.Atom("H1", MM.Element("H"), 8),
                             MM.Atom("H2", MM.Element("H"), 8),
                             MM.Atom("O",  MM.Element("O"), 2)),
                            (("H1", "O", "single"), ("H2", "O", "single")))
        return MM.Universe('infinite', [(water, 10)],
                           convention='my_own')

    def make_immutable_universe(self):
        water = IM.fragment("water", (),
                            (("H1", IM.atom(IM.element("H"), 8)),
                             ("H2", IM.atom(IM.element("H"), 8)),
                             ("O",  IM.atom(IM.element("O"), 2))),
                            (("H1", "O", "single"), ("H2", "O", "single")))
        return IM.universe('infinite', [(water, "solvent", 10)],
                           convention='my_own')

    def test_equivalence(self):
        models = [self.make_mutable_universe(),
                  self.make_immutable_universe()]
        models.extend([AM.Universe(m) for m in models])
        for m1 in models:
            for m2 in models:
                self.assertTrue(m1.is_equivalent(m2))


class PeptideTest(unittest.TestCase):

    def make_mutable_universe(self):
        C = MM.Element('C')
        H = MM.Element('H')
        N = MM.Element('N')
        O = MM.Element('O')
        peptide_group = MM.Fragment('peptide', 'peptide',
                                    (),
                                    (MM.Atom('CA', C),
                                     MM.Atom('HA', H),
                                     MM.Atom('H', H),
                                     MM.Atom('N', N),
                                     MM.Atom('C', C),
                                     MM.Atom('O', O)),
                                    (('N', 'H', "single"),
                                     ('N', 'CA', "single"),
                                     ('CA', 'HA', "single"),
                                     ('CA', 'C', "single"),
                                     ('C', 'O', "double")))
        ala_sidechain = MM.Fragment('sidechain', 'ala_sidechain',
                                    (),
                                    (MM.Atom('CB', C),
                                     MM.Atom('HB1', H),
                                     MM.Atom('HB2', H),
                                     MM.Atom('HB3', H)),
                                    (('CB', 'HB1', "single"),
                                     ('CB', 'HB2', "single"),
                                     ('CB', 'HB3', "single"),))
        ala = lambda label: MM.Fragment(label, 'alanine',
                                        (copy.copy(peptide_group),
                                         copy.copy(ala_sidechain)),
                                        (),
                                        (('peptide.CA', 'sidechain.CB',
                                          "single"),))
        di_ala = MM.Polymer('di_ala', 'alanine_dipeptide',
                            (ala('ALA1'), ala('ALA2')),
                            (('ALA1.peptide.C', 'ALA2.peptide.N', "single"),),
                            'polypeptide')
        water = MM.Fragment("solvent", "water", (),
                            (MM.Atom("H1", MM.Element("H"), 8),
                             MM.Atom("H2", MM.Element("H"), 8),
                             MM.Atom("O",  MM.Element("O"), 2)),
                            (("H1", "O", "single"), ("H2", "O", "single")))
        return MM.Universe('cube', [(di_ala, 1), (water, 10)],
                           convention='my_own')

    def make_immutable_universe(self):
        C = IM.element('C')
        H = IM.element('H')
        N = IM.element('N')
        O = IM.element('O')
        peptide_group = IM.fragment('peptide',
                                    (),
                                    (('CA', IM.atom(C)),
                                     ('HA', IM.atom(H)),
                                     ('H', IM.atom(H)),
                                     ('N', IM.atom(N)),
                                     ('C', IM.atom(C)),
                                     ('O', IM.atom(O))),
                                    (('N', 'H', "single"),
                                     ('N', 'CA', "single"),
                                     ('CA', 'HA', "single"),
                                     ('CA', 'C', "single"),
                                     ('C', 'O', "double")))
        ala_sidechain = IM.fragment('ala_sidechain',
                                    (),
                                    (('CB', IM.atom(C)),
                                     ('HB1', IM.atom(H)),
                                     ('HB2', IM.atom(H)),
                                     ('HB3', IM.atom(H))),
                                    (('CB', 'HB1', "single"),
                                     ('CB', 'HB2', "single"),
                                     ('CB', 'HB3', "single"),))
        ala = IM.fragment('alanine',
                          (('peptide', peptide_group),
                           ('sidechain', ala_sidechain)),
                          (),
                          (('peptide.CA', 'sidechain.CB', "single"),))
        di_ala = IM.polymer('alanine_dipeptide',
                            (('ALA1', ala),
                             ('ALA2', ala)),
                            (('ALA1.peptide.C', 'ALA2.peptide.N', "single"),),
                            'polypeptide')
        water = IM.fragment("water", (),
                            (("H1", IM.atom(H, 8)),
                             ("H2", IM.atom(H, 8)),
                             ("O",  IM.atom(O, 2))),
                            (("H1", "O", "single"), ("H2", "O", "single")))
        return IM.universe('cube',
                           [(di_ala, "di_ala", 1),
                            (water, "solvent", 10)],
                           convention='my_own')

    def make_configuration(self, M, universe):
        return M.Configuration(universe,
                               IN.zeros((universe.number_of_sites, 3),
                                        N.float32),
                               IN.array(1, N.float32))

    def make_masses(self, M, universe):
        mass = {'C': 12., 'H': 1., 'N': 14., 'O': 16.}
        m = [mass[a.name]
             for a in IT.chain.from_iterable(f.recursive_atom_iterator()
                                             for f, _ in universe.molecules)]
        m = IN.array(m, N.float32)
        try:
            return M.TemplateAtomProperty(universe, "mass", "g mol-1", m)
        except AttributeError:
            return M.Property(universe, "mass", "g mol-1", m, 'template_atom')

    def test_equivalence(self):
        models = [self.make_immutable_universe(), self.make_mutable_universe()]
        models.extend([AM.Universe(m) for m in models])
        for m1 in models:
            for m2 in models:
                self.assertTrue(m1.is_equivalent(m2))
        confs = [self.make_configuration(M, u)
                 for M, u in zip((IM, MM, AM), models)]
        for c1 in confs:
            for c2 in confs:
                self.assertTrue(c1.is_equivalent(c2))
        props = [self.make_masses(M, u)
                 for M, u in zip((IM, MM, AM), models)]
        for p1 in props:
            for p2 in props:
                self.assertTrue(p1.is_equivalent(p2))

    def test_factory(self):
        models = [self.make_immutable_universe(),
                  self.make_mutable_universe()]
        models.extend([AM.Universe(m) for m in models])
        confs = [self.make_configuration(M, u)
                 for M, u in zip((IM, MM, AM), models)]
        factories = [M.Factory() for M in (IM, MM, AM)]
        for f in factories:
            for c in confs:
                c.validate()
                self.assertTrue(is_valid(c))
                fc = f(c)
                self.assertTrue(c.is_equivalent(fc))
                self.assertTrue(is_valid(fc))

def suite():
    loader = unittest.TestLoader()
    s = unittest.TestSuite()
    s.addTest(loader.loadTestsFromTestCase(WaterTest))
    s.addTest(loader.loadTestsFromTestCase(PeptideTest))
    return s

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# Tests for module mosaic.array_model

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import unittest

import numpy as N

import mosaic.mutable_model as M
import mosaic.array_model as AM
from mosaic.api import is_valid

class UniverseTest(unittest.TestCase):

    def setUp(self):
        self.factory = AM.Factory()
        mol = M.Fragment("water", "water", (),
                         (M.Atom("H1", M.Element("H")),
                          M.Atom("H2", M.Element("H")),
                          M.Atom("O",  M.Element("O"), 2)),
                         (("H1", "O", "single"), ("H2", "O", "single")))
        self.universe = M.Universe('infinite', [(mol, 10)],
                                   convention='my_own')
        self.am_universe = self.factory(self.universe)

    def test_basics(self):
        self.assertTrue(self.am_universe.is_equivalent(self.universe))
        self.assertEqual(self.am_universe.number_of_molecules, 10)
        self.assertEqual(self.am_universe.number_of_atoms, 30)
        self.assertEqual(self.am_universe.number_of_sites, 40)
        self.assertEqual(self.am_universe.number_of_bonds, 20)
        self.assertEqual(self.am_universe.cell_shape, "infinite")
        self.assertEqual(self.am_universe.convention, "my_own")

    def test_properties(self):
        mm_masses = M.TemplateAtomProperty(self.universe, "mass", "amu",
                                           N.array([1., 1., 16.], N.float32))
        masses = self.factory(mm_masses)
        self.assertTrue(masses.is_equivalent(mm_masses))
        self.assertEqual(masses.type, 'template_atom')
        self.assertTrue(masses.universe is self.am_universe)
        self.assertEqual(masses.element_shape, ())
        self.assertEqual(masses.data.shape, (3,))
        mm_masses = M.TemplateSiteProperty(self.universe, "mass", "amu",
                                           N.array([1., 1., 8., 8.], N.float32))
        bead_masses = self.factory(mm_masses)
        self.assertTrue(bead_masses.is_equivalent(mm_masses))
        self.assertEqual(bead_masses.type, 'template_site')
        self.assertTrue(bead_masses.universe is self.am_universe)
        self.assertEqual(bead_masses.element_shape, ())
        self.assertEqual(bead_masses.data.shape, (4,))
        mm_velocities = M.SiteProperty(self.universe, "velocity", "nm ps-1",
                                       dtype=N.float64, element_shape=(3,))
        velocities = self.factory(mm_velocities)
        self.assertTrue(velocities.is_equivalent(mm_velocities))
        self.assertEqual(velocities.type, 'site')
        self.assertTrue(velocities.universe is self.am_universe)
        self.assertEqual(velocities.data.shape, (40, 3))
        self.assertEqual(velocities.element_shape, (3,))
        mm_ap = M.AtomProperty(self.universe, "foo", "",
                               dtype=N.int16, element_shape=(2, 2))
        ap = self.factory(mm_ap)
        self.assertTrue(ap.is_equivalent(mm_ap))
        self.assertEqual(ap.type, 'atom')
        self.assertTrue(ap.universe is self.am_universe)
        self.assertEqual(ap.data.shape, (30, 2, 2))
        self.assertEqual(ap.element_shape, (2, 2))

    def test_labels(self):
        labels = tuple(a.name
                       for f, n in self.universe.molecules
                       for a in f.recursive_atom_iterator())
        mm_el = M.TemplateAtomLabel(self.universe, "element", labels)
        el = self.factory(mm_el)
        self.assertTrue(is_valid(el))
        self.assertTrue(el.is_equivalent(mm_el))
        self.assertEqual(el.name, "element")
        self.assertTrue(el.universe.is_equivalent(self.universe))
        self.assertTrue(len(el.strings)
                        == self.universe.number_of_template_atoms)
        for s1, s2 in zip(labels, el.strings):
            self.assertEqual(s1, s2)

        labels = tuple(a.name
                       for f, n in self.universe.molecules
                       for _ in range(n)
                       for a in f.recursive_atom_iterator())
        mm_el = M.AtomLabel(self.universe, "element", labels)
        el = self.factory(mm_el)
        self.assertTrue(is_valid(el))
        self.assertTrue(el.is_equivalent(mm_el))
        self.assertEqual(el.name, "element")
        self.assertTrue(el.universe.is_equivalent(self.universe))
        self.assertTrue(len(el.strings)
                        == self.universe.number_of_atoms)
        for s1, s2 in zip(labels, el.strings):
            self.assertEqual(s1, s2)

        labels = tuple(a.name
                       for f, n in self.universe.molecules
                       for a in f.recursive_atom_iterator()
                       for _ in range(a.number_of_sites))
        mm_el = M.TemplateSiteLabel(self.universe, "element", labels)
        el = self.factory(mm_el)
        self.assertTrue(is_valid(el))
        self.assertTrue(el.is_equivalent(mm_el))
        self.assertEqual(el.name, "element")
        self.assertTrue(el.universe.is_equivalent(self.universe))
        self.assertTrue(len(el.strings)
                        == self.universe.number_of_template_sites)
        for s1, s2 in zip(labels, el.strings):
            self.assertEqual(s1, s2)

        labels = tuple(a.name
                       for f, n in self.universe.molecules
                       for _ in range(n)
                       for a in f.recursive_atom_iterator()
                       for __ in range(a.number_of_sites))
        mm_el = M.SiteLabel(self.universe, "element", labels)
        el = self.factory(mm_el)
        self.assertTrue(is_valid(el))
        self.assertTrue(el.is_equivalent(mm_el))
        self.assertEqual(el.name, "element")
        self.assertTrue(el.universe.is_equivalent(self.universe))
        self.assertTrue(len(el.strings)
                        == self.universe.number_of_sites)
        for s1, s2 in zip(labels, el.strings):
            self.assertEqual(s1, s2)
        
    def test_bonds(self):
        bonds = self.universe.bond_index_array()
        self.assertEqual(len(bonds), self.universe.number_of_bonds)
        self.assertTrue((bonds >= 0).all())
        self.assertTrue((bonds < self.universe.number_of_atoms).all())
        for i in range(10):
            self.assertEqual(bonds[2*i, 0], 3*i)
            self.assertEqual(bonds[2*i, 1], 3*i+2)
            self.assertEqual(bonds[2*i+1, 0], 3*i+1)
            self.assertEqual(bonds[2*i+1, 1], 3*i+2)

    def test_selections(self):
        universe = self.am_universe
        s = AM.Selection(universe, N.array([0, 2], N.uint8), 'atom')
        self.assertEqual(s.number_of_atoms, 2)
        self.assertEqual(s.number_of_sites, 3)
        s = AM.Selection(universe, N.array([1], N.uint8), 'template_atom')
        self.assertEqual(s.number_of_atoms, 10)
        self.assertEqual(s.number_of_sites, 10)
        s = AM.Selection(universe, N.array([2], N.uint8), 'template_atom')
        self.assertEqual(s.number_of_atoms, 10)
        self.assertEqual(s.number_of_sites, 20)
        s = AM.Selection(universe, N.array([2], N.uint8), 'site')
        self.assertEqual(s.number_of_sites, 1)
        s = AM.Selection(universe, N.array([0], N.uint8), 'template_site')
        self.assertEqual(s.number_of_sites, 10)

    def test_optimizations(self):
        import mosaic.api
        universe = self.am_universe
        for a1, a2 in zip(universe.recursive_atom_iterator(),
                          mosaic.api.MosaicUniverse.recursive_atom_iterator(universe)):
            self.assertTrue(a1.is_equivalent(a2))
        bonds1 = universe.bond_index_array()
        bonds2 = mosaic.api.MosaicUniverse.bond_index_array(universe)
        self.assertTrue((bonds1 == bonds2).all())

def suite():
    loader = unittest.TestLoader()
    s = unittest.TestSuite()
    s.addTest(loader.loadTestsFromTestCase(UniverseTest))
    return s

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""PDB Chemical Component Dictionary

The PDB uses a dictionary defining all the chemical components from
its structure files, each identified by a three-letter code. This
dictionary permits in particular the identification of chemical bonds.
The dictionary consists of a huge mmCIF file plus a smaller one.
Downloading and parsing these files takes a lot of time, therefore
the result is cached in a pickle file in $HOME/.mosaic.

The dictionary is updated as new components are added with new
structures being published. If a conversion fails, it's a good idea to
delete the file
$HOME/.mosaic/pdb_chemical_components_dictionary.pickle which is
then regenerated automatically on the next run.

.. moduleauthor:: Konrad Hinsen <konrad.hinsen@cnrs-orleans.fr>

"""

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import gzip
import os
import sys

# Python 2/3 compatibility issues
if sys.version_info[0] == 2:
    from StringIO import StringIO
    from urllib2 import urlopen
    import cPickle as pickle
    bytes2text = lambda x: x
else:
    from io import BytesIO as StringIO
    from urllib.request import urlopen
    import pickle
    from io import TextIOWrapper
    def bytes2text(stream):
        return TextIOWrapper(stream, encoding="utf8")

from mosaic_pdb import mmcif

# For now, MOSAIC_DIR is used only for the PDB Chemical Components Dictionary,
# which is why it is handled here.

MOSAIC_DIR = os.path.join(os.environ['HOME'], '.mosaic')
if not os.path.exists(MOSAIC_DIR):
    os.makedirs(MOSAIC_DIR)

# PDB Chemical Components Dictionary
#
# ftp://ftp.wwpdb.org/pub/pdb/data/monomers/components.cif.gz
# ftp://ftp.wwpdb.org/pub/pdb/data/monomers/aa-variants-v1.cif.gz
#
# These big files are downloaded once and converted to a pickle
# file stored in PDB_CHEM_COMP_DICT that is loaded much faster.

if sys.version_info[0] == 2:
    PDB_CHEM_COMP_DICT = os.path.join(MOSAIC_DIR,
                                      'pdb_chemical_components_dictionary.pickle')
else:
    PDB_CHEM_COMP_DICT = os.path.join(MOSAIC_DIR,
                                      'pdb_chemical_components_dictionary_py3.pickle')

chem_comp_urls = ['ftp://ftp.wwpdb.org/pub/pdb/data/monomers/'
                  'components.cif.gz',
                  'ftp://ftp.wwpdb.org/pub/pdb/data/monomers/'
                  'aa-variants-v1.cif.gz']


# The in-memory representation of the dictionary

class ChemComp(object):

    def __repr__(self):
        return "ChemComp('%s')" % self.id


class Table(object):

    def __init__(self, name, header, conversions):
        self.name = name
        self.fields = [f for f in header if f in conversions]
        self.select = [(i, conversions[f])
                       for i, f in enumerate(header)
                       if f in conversions]
        self.entries = []

    def addEntry(self, data):
        self.entries.append(tuple(c(data[i]) for i, c in self.select))

    def __repr__(self):
        return "Table('%s')" % self.name

    def __getitem__(self, item):
        if isinstance(item, int):
            data = self.entries[item]
            return dict((k, data[i]) for i, k in enumerate(self.fields))
        elif isinstance(item, str):
            try:
                i = self.fields.index(item)
            except ValueError:
                raise KeyError(item)
            return [e[i] for e in self.entries]


class SingleEntryTable(Table):

    def __init__(self, name, conversions):
        self.name = name
        self.conversions = conversions
        self.fields = []
        self.entries = [[]]

    def __setitem__(self, item, value):
        if item in self.conversions:
            self.fields.append(item)
            self.entries[0].append(self.conversions[item](value))


def string_item(s):
    if s == '?':
        return None
    else:
        return s


def float_item(s):
    if s == '?':
        return None
    else:
        return float(s)


# Try to load the pickle file. If it doesn't exist, load
# the data from the official URLs and create the pickle file.

try:

    with open(PDB_CHEM_COMP_DICT, 'rb') as pickled_dict:
        components, variants = pickle.load(pickled_dict)

except IOError:

    sys.stderr.write("Downloading the PDB Chemical Component Dictionary.\n")
    sys.stderr.write("This make take a few minutes.\n")

    table_data = {'chem_comp_atom': 'atoms',
                  'chem_comp_bond': 'bonds'}

    table_items = {'chem_comp_atom': {'atom_id': string_item,
                                      'type_symbol': string_item,
                                      'charge': float_item},
                   'chem_comp_bond': {'atom_id_1': string_item,
                                      'atom_id_2': string_item,
                                      'value_order': string_item}}

    components = {}
    variants = {}

    for url in chem_comp_urls:
        buffer = StringIO(urlopen(url).read())
        f = bytes2text(gzip.GzipFile(fileobj=buffer))
        parser = mmcif.MMCIFParser(file_object=f)
        for item_type, item in parser.parse():
            if item_type is mmcif.KEYWORD:
                if item[0] == 'data':
                    comp_id = item[1]
                    comp = ChemComp()
                    components[comp_id] = comp
                else:
                    raise ValueError("Keyword %s not yet implemented"
                                     % item[0])
            elif item_type is mmcif.TABLE_HEADER:
                table_name = item[0][0]
                if table_name in table_data:
                    table = Table(table_name, [i[1] for i in item],
                                  table_items[table_name])
                    setattr(comp, table_data[table_name], table)
                else:
                    table = None
            elif item_type is mmcif.TABLE_DATA:
                if table is not None:
                    table.addEntry(item)
            elif item_type is mmcif.DATA:
                if item[0] == 'chem_comp':
                    setattr(comp, item[1], item[2])
                    if item[1] == 'three_letter_code':
                        if item[2] != comp_id:
                            vs = variants.get(item[2], [])
                            vs.append(comp_id)
                            variants[item[2]] = vs
                elif item[0] in ['chem_comp_atom', 'chem_comp_bond']:
                    try:
                        table = getattr(comp, table_data[item[0]])
                    except AttributeError:
                        table = SingleEntryTable(item[0], table_items[item[0]])
                        setattr(comp, table_data[item[0]], table)
                    table[item[1]] = item[2]
                else:
                    if item[0] not in ['pdbx_chem_comp_descriptor',
                                       'pdbx_chem_comp_identifier',
                                       'pdbx_chem_comp_feature',
                                       'pdbx_chem_comp_audit']:
                        raise ValueError("Unexpected data item %s" % item[0])
            else:
                raise ValueError("Unexpected item type %s"
                                 % str(item_type),
                                 parser.line_number)

    f = open(PDB_CHEM_COMP_DICT, 'wb')
    pickle.dump((components, variants), f, protocol=-1)
    f.close()



# -*- coding: utf-8 -*-
"""Immutable model

.. moduleauthor:: Konrad Hinsen

This experimental module implements the Mosaic API in terms of
immutable data structures as implemented in ImmutablePy.  The use of
immutable data structures avoids many nasty bugs, such as modifying a
universe after creating a configuration object that refer to it,
invalidating the configuration. However, the immutable data structures
from ImmutablePy are not very efficient at the moment.

"""

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import copy
import collections
import itertools as IT
import types

import numpy as N

from immutable import Immutable, ImmutableTuple, ImmutableDict, immutable
import immutable.np as IN

import mosaic.api as api
from mosaic.utility import AbstractFactory
from mosaic.utility import MethodRegister
from mosaic.utility import isascii
from mosaic.utility import uint_for_max_value


# Look up template objects in the class' instance cache.
# If not found, create new object.

def _lookup(klass, key):
    key = immutable(key)
    try:
        # Under Python 3, this raises a TypeError
        # for ImmutableNDArray.
        return klass._instances[key]
    except (KeyError, TypeError):
        obj = klass(key)
        try:
            klass._instances[key] = obj
        except TypeError:
            pass
        return obj


# Atom descriptors

class AtomDescriptor(Immutable):

    def __init__(self, name=""):
        self._validate_name(name)
        self._name = name

    def _validate_name(cls, name):
        api.validate_label(name, "name")

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self._name)

    @property
    def name(self):
        return self._name


class Element(AtomDescriptor):

    _instances = {}

    def _validate_name(cls, name):
        api.MosaicAtom.validate_element_name(name)

    def __repr__(self):
        return "Element('%s')" % self._name

    # Data Model API implementation
    @property
    def type(self):
        return "element"


class CGParticle(AtomDescriptor):

    _instances = {}

    def __repr__(self):
        if self._name:
            return "CGParticle('%s')" % self._name
        else:
            return "CGParticle()"

    @property
    def type(self):
        return "cgparticle"


class Dummy(AtomDescriptor):

    _instances = {}

    def __repr__(self):
        if self._name:
            return "Dummy('%s')" % self._name
        else:
            return "Dummy()"

    @property
    def type(self):
        return "dummy"


class Unknown(AtomDescriptor):

    _instances = {}

    def __repr__(self):
        if self._name:
            return "Unknown('%s')" % self._name
        else:
            return "Unknown()"

    @property
    def type(self):
        return ""


element = lambda name: _lookup(Element, name)
cgparticle = lambda name: _lookup(CGParticle, name)
dummy = lambda name="": _lookup(Dummy, name)
unknown = lambda name="": _lookup(Unknown, name)


# Atom and fragment templates

class Atom(Immutable):

    _instances = {}

    def __init__(self, args):
        descriptor, nsites = args
        api.validate_type(descriptor, AtomDescriptor, "atom descriptor")
        self.descriptor = descriptor
        api.validate_type(nsites, int, "number of sites")
        if nsites < 1:
            raise ValueError("number of sites must be >= 1")
        self.nsites = nsites

    def __repr__(self):
        return "Atom(%s, %d)" % (repr(self.descriptor), self.nsites)

    # API-compatible properties

    @property
    def number_of_atoms(self):
        return 1

    @property
    def number_of_sites(self):
        return self.nsites

class AtomRef(Immutable, api.MosaicAtom):

    def __init__(self, label, atom):
        self._label = label
        self._atom = atom

    # Data Model API implementation

    @property
    def label(self):
        return self._label

    @property
    def name(self):
        return self._atom.descriptor.name

    @property
    def type(self):
        return self._atom.descriptor.type

    @property
    def number_of_sites(self):
        return self._atom.nsites

atom = lambda descriptor, nsites=1: _lookup(Atom, (descriptor, nsites))


class Fragment(Immutable, collections.Mapping):

    _instances = {}

    def __init__(self, args):
        species, fragments, atoms, bonds = args
        api.validate_label(species, "species")
        self.species = species

        labels = set()

        api.validate_sequence(fragments, ImmutableTuple, "fragments",
                              ((lambda p: len(p) == 2, "must have length 2"),))
        for label, fragment in fragments:
            api.validate_label(label, "fragment label")
            if label in labels:
                raise ValueError("label %s occurs more than once" % label)
            labels.add(label)
            api.validate_type(fragment, Fragment, "fragment template")
        self.fragments = fragments

        api.validate_sequence(atoms, ImmutableTuple, "atoms",
                              ((lambda p: len(p) == 2, "must have length 2"),))
        for label, atom in atoms:
            api.validate_label(label, "atom label")
            if label in labels:
                raise ValueError("label %s occurs more than once" % label)
            labels.add(label)
            api.validate_type(atom, Atom, "atom")
        self.atoms = atoms

        self.attrs = ImmutableDict(IT.chain(fragments, atoms))

        api.validate_sequence(bonds, tuple, "bonds",
                              ((lambda p: len(p) == 3, "must have length 3"),))
        for a1, a2, order in bonds:
            self._validate_bond_atom(a1)
            self._validate_bond_atom(a2)
            if a1.split('.')[0] == a2.split('.')[0]:
                raise ValueError("bond between %s and %s must be defined "
                                 "in fragment %s" % (a1, a2, a1.split('.')[0]))
            api.validate_value(order, api.MosaicFragment._bond_orders,
                               "bond order")
        self.bonds = bonds

    def _validate_bond_atom(self, atom_label):
        api.validate_type(atom_label, str, "atom label")
        obj = self
        for item in atom_label.split('.'):
            if isinstance(obj, Atom):
                raise ValueError("invalid atom reference "
                                 "%s (refers to child object of an atom)"
                                 % atom_label)
            try:
                obj = obj.attrs[item]
            except KeyError:
                raise ValueError("invalid atom reference %s "
                                 "(child %s not found)" % (atom_label, item))

    def __repr__(self):
        return "Fragment(%s, ...)" % self.species

    def __eq__(self, other):
        return isinstance(other, Fragment) \
               and self.species == other.species \
               and all(s == o
                       for s, o in zip(self.fragments, other.fragments)) \
               and all(s == o
                       for s, o in zip(self.atoms, other.atoms)) \
               and frozenset([(frozenset([a1, a2]), o)
                              for a1, a2, o in self.bonds]) == \
                   frozenset([(frozenset([a1, a2]), o)
                              for a1, a2, o in other.bonds])

    def __hash__(self):
        # Don't use bonds in hash because the bond list can be
        # different for fragments that test equal.
        return hash((self.species, self.fragments, self.atoms))

    # Mapping interface

    def __getitem__(self, item):
        return self.attrs[item]

    def __len__(self):
        return len(self.attrs)

    def __iter__(self):
        return iter(self.attrs)

    # API-compatible properties

    @property
    def number_of_atoms(self):
        return sum(f.number_of_atoms for l, f in self.fragments) \
               + len(self.atoms)

    @property
    def number_of_sites(self):
        return sum(f.number_of_sites for l, f in self.fragments) \
               + sum(a.number_of_sites for l, a in self.atoms)

    @property
    def number_of_bonds(self):
        return sum(f.number_of_bonds for l, f in self.fragments) \
               + len(self.bonds)

class Polymer(Fragment, collections.Sequence):

    _instances = {}

    def __init__(self, args):
        species, fragments, bonds, polymer_type = args
        Fragment.__init__(self, (species, fragments, ImmutableTuple(), bonds))
        self.polymer_type = polymer_type

    def __repr__(self):
        return "Polymer(%s, ...)" % self.species

    # Sequence and mapping interface
    def __getitem__(self, item):
        if isinstance(item, int):
            return self.fragments[int]
        else:
            return Fragment.__getitem__(self, item)

class FragmentRef(Immutable, api.MosaicFragment):

    def __init__(self, label, fragment):
        self._label = label
        self._fragment = fragment

    # Data Model API implementation

    @property
    def label(self):
        return self._label

    @property
    def species(self):
        return self._fragment.species

    @property
    def fragments(self):
        return [FragmentRef(l, f) for l, f in self._fragment.fragments]

    @property
    def atoms(self):
        return [AtomRef(l, a) for l, a in self._fragment.atoms]

    @property
    def bonds(self):
        return self._fragment.bonds

    @property
    def is_polymer(self):
        return isinstance(self._fragment, Polymer)

    @property
    def polymer_type(self):
        if isinstance(self._fragment, Polymer):
            return self._fragment.polymer_type
        else:
            raise ValueError("polymer_type not defined in a non-polymer")

    # Sequence and mapping interface

    def __getitem__(self, item):
        return self._fragment[item]

fragment = lambda species, fragments, atoms, bonds: \
            _lookup(Fragment, (species, fragments, atoms, bonds))

polymer = lambda species, fragments, bonds, polymer_type: \
            _lookup(Polymer, (species, fragments, bonds, polymer_type))


# Universe

class Universe(Immutable, api.MosaicUniverse):

    _instances = {}

    def __init__(self, args):
        cell_shape, molecules, symmetry_transformations, convention \
                = args
        api.validate_type(cell_shape, str, "cell_shape")
        api.validate_value(cell_shape,
                           list(self._cell_parameter_array_shapes.keys()),
                           "cell_shape")
        api.validate_sequence(molecules, ImmutableTuple, "molecules",
                              ((lambda p: len(p) == 3, "must have length 3"),
                               (lambda p: isinstance(p[0], Fragment)
                                          and isascii(p[1])
                                          and isinstance(p[2], int),
                                "elements must be (fragment, label, count) "
                                "triples")))
        api.validate_sequence(symmetry_transformations, ImmutableTuple,
                              "symmetry_transformations",
                              ((lambda p: len(p) == 2, "must have length 2"),
                               (lambda p: hasattr(p[0], 'shape')
                                          and p[0].shape == (3, 3)
                                          and p[0].dtype == N.float64,
                                "rotation matrix must be float64 "
                                "and have shape (3,3)"),
                               (lambda p: hasattr(p[1], 'shape')
                                          and p[1].shape == (3,)
                                          and p[0].dtype == N.float64,
                                "translation vector must be float64 "
                                "and have shape (3,)"),))
        if cell_shape == "infinite" \
           and len(symmetry_transformations) > 0:
            raise ValueError("Symmetry transformations are allowed "
                             "only in periodic universes")
        api.validate_label(convention, "Universe.convention")

        self._cell_shape = cell_shape
        self._molecules = ImmutableTuple(
                ImmutableTuple((FragmentRef(label, fragment), count))
                for fragment, label, count in molecules)
        self._symmetry_transformations = symmetry_transformations
        self._convention = convention

        self._fragments = ImmutableTuple(ImmutableTuple((f, l))
                                         for f, l, c in molecules)
        self._molecule_counts = ImmutableTuple(c for f, l, c in molecules)
        self._atom_counts = ImmutableTuple(f.number_of_atoms
                                           for f, l, c in molecules)
        self._site_counts = ImmutableTuple(f.number_of_sites
                                           for f, l, c in molecules)
        self._bond_counts = ImmutableTuple(f.number_of_bonds
                                           for f, l, c in molecules)

    def __eq__(self, other):
        return isinstance(other, Universe) \
               and self.cell_shape == other.cell_shape \
               and self.symmetry_transformations == \
                       other.symmetry_transformations \
               and self.convention == other.convention \
               and all(c1 == c2
                       for c1, c2 in zip(self._molecule_counts,
                                         other._molecule_counts)) \
               and all(f1 == f2
                       for f1, f2 in zip(self._fragments, other._fragments))

    def __hash__(self):
        # a cheap hash that doesn't look at the fragments
        return hash(self._cell_shape) + len(self._fragments)

    def __repr__(self):
        mol_str = ', '.join("(Fragment('%s'), %d)" % (f.label, c)
                            for f, c in self._molecules)
        return "Universe('%s', [%s])" % (self.cell_shape, mol_str)

    # Data Model API implementation

    @property
    def cell_shape(self):
        return self._cell_shape

    @property
    def symmetry_transformations(self):
        return self._symmetry_transformations

    @property
    def convention(self):
        return self._convention

    @property
    def molecules(self):
        return self._molecules

    # More efficient implementations of the number properties

    @property
    def number_of_molecules(self):
        return sum(self._molecule_counts)

    @property
    def number_of_atoms(self):
        return sum(c * na for c, na in zip(self._molecule_counts,
                                           self._atom_counts))

    @property
    def number_of_sites(self):
        return sum(c * ns for c, ns in zip(self._molecule_counts,
                                           self._site_counts))

    @property
    def number_of_bonds(self):
        return sum(c * nb for c, nb in zip(self._molecule_counts,
                                           self._bond_counts))

    @property
    def number_of_template_atoms(self):
        return sum(self._atom_counts)

    @property
    def number_of_template_sites(self):
        return sum(self._site_counts)

universe = lambda cell_shape, molecules, \
                    symmetry_transformations=(), convention='': \
            _lookup(Universe, (cell_shape, molecules,
                               symmetry_transformations, convention))


# Properties

class Property(Immutable, api.MosaicProperty):

    def __init__(self, universe, name, units, data):
        api.validate_type(universe, Universe, "universe")
        self._universe = universe
        api.validate_label(name, "name")
        self._name = name
        api.validate_units(units, "units")
        self._units = units
        api.validate_array(data, None, self._allowed_dtypes, "data")
        if data.shape[0] != self._number_of_values():
            raise ValueError("data array has incorrect shape")
        self._data = data

    def __repr__(self):
        return "%s()" % (self.__class__.__name__)

    def __eq__(self, other):
        return isinstance(other, Property) \
               and self.type == other.type \
               and self.name == other.name \
               and self.units == other.units \
               and self.universe == other.universe \
               and (self.data == other.data).all()

    def __hash__(self):
        # cheap hash function that doesn't look at the data
        return hash((self.type, self.universe, self.data.shape))

    # Data Model API implementation
    @property
    def universe(self):
        return self._universe

    @property
    def name(self):
        return self._name

    @property
    def units(self):
        return self._units

    @property
    def data(self):
        return self._data


class AtomProperty(Property):

    def _number_of_values(self):
        return self._universe.number_of_atoms

    # Data Model API implementation
    @property
    def type(self):
        return 'atom'


class SiteProperty(Property):

    def _number_of_values(self):
        return self._universe.number_of_sites

    # Data Model API implementation
    @property
    def type(self):
        return 'site'


class TemplateAtomProperty(Property):

    def _number_of_values(self):
        return self._universe.number_of_template_atoms

    # Data Model API implementation
    @property
    def type(self):
        return 'template_atom'


class TemplateSiteProperty(Property):

    def _number_of_values(self):
        return self._universe.number_of_template_sites

    # Data Model API implementation
    @property
    def type(self):
        return 'template_site'


# Labels

class Label(Immutable, api.MosaicLabel):

    def __init__(self, universe, name, strings):
        api.validate_type(universe, Universe, "universe")
        self._universe = universe
        api.validate_label(name, "name")
        self._name = name
        api.validate_sequence(strings, str, "strings")
        if len(strings) != self._number_of_values():
            raise ValueError("incorrect number of strings")
        for s in strings:
            api.validate_ascii_string(s, "label")
        self._strings = ImmutableTuple(strings)

    def __repr__(self):
        return "%s()" % (self.__class__.__name__)

    def __eq__(self, other):
        return isinstance(other, Label) \
               and self.type == other.type \
               and self.name == other.name \
               and self.universe == other.universe \
               and self.strings == other.strings

    def __hash__(self):
        return hash((self.type, self.name, self.universe, self.strings))

    # Data Model API implementation
    @property
    def universe(self):
        return self._universe

    @property
    def name(self):
        return self._name

    @property
    def strings(self):
        return self._strings


class AtomLabel(Label):

    def _number_of_values(self):
        return self._universe.number_of_atoms

    # Data Model API implementation
    @property
    def type(self):
        return 'atom'


class SiteLabel(Label):

    def _number_of_values(self):
        return self._universe.number_of_sites

    # Data Model API implementation
    @property
    def type(self):
        return 'site'


class TemplateAtomLabel(Label):

    def _number_of_values(self):
        return self._universe.number_of_template_atoms

    # Data Model API implementation
    @property
    def type(self):
        return 'template_atom'


class TemplateSiteLabel(Label):

    def _number_of_values(self):
        return self._universe.number_of_template_sites

    # Data Model API implementation
    @property
    def type(self):
        return 'template_site'


# Configuration

class Configuration(Immutable, api.MosaicConfiguration):

    def __init__(self, universe, positions, cell_parameters):
        api.validate_type(universe, Universe, "universe")
        cell_param_shape = universe.cell_parameter_array_shape
        nsites = universe.number_of_sites
        # Allow cell_parameters=None for universes that don't
        # need cell parameters.
        if cell_parameters is None and cell_param_shape == (0,):
            cell_parameters = IN.zeros(cell_param_shape,
                                       positions.dtype)
        # At this point, positions and cell_parameters must be arrays
        # of the required shapes.
        api.validate_array(positions, (nsites, 3),
                           self._allowed_dtypes, "positions")
        api.validate_array(cell_parameters, cell_param_shape,
                           self._allowed_dtypes, "cell_parameters")
        if positions.dtype != cell_parameters.dtype:
            raise ValueError("positions and cell parameters must have"
                             " the same element type")

        self._universe = universe
        self._positions = positions
        self._cell_parameters = cell_parameters

    def __repr__(self):
        return "Configuration()"

    def __eq__(self, other):
        return isinstance(other, Configuration) \
               and self.universe == other.universe \
               and (self.cell_parameters == other.cell_parameters).all() \
               and (self.positions == other.positions).all()

    def __hash__(self):
        # cheap hash function that doesn't look at the positions
        return hash((self.universe, self.positions.shape,
                     tuple(self.cell_parameters.flat)))

    # Reimplementation that returns immutable data
    # Inactive for now because this leads to strange bugs with various
    # versions of NumPy.
    if False:
        def lattice_vectors(self):
            return immutable(api.MosaicConfiguration.lattice_vectors(self))

    # Data Model API implementation
    @property
    def universe(self):
        return self._universe

    @property
    def cell_parameters(self):
        return self._cell_parameters

    @property
    def positions(self):
        return self._positions


class Selection(Immutable, api.MosaicSelection):

    def __init__(self, universe, indices):
        api.validate_type(universe, Universe, "universe")
        self._universe = universe
        indices = IN.array(indices, uint_for_max_value(max(indices)))
        api.validate_indices(indices, self._max_index(), "indices")
        self._indices = indices

    def __str__(self):
        return "%s()" % (self.__class__.__name__)

    def __eq__(self, other):
        return isinstance(other, Selection) \
               and self.type == other.type \
               and self.universe == other.universe \
               and (self.indices == other.indices).all()

    def __hash__(self):
        return hash(self.type) + hash(self.universe) + hash(len(self.indices))

    # Data Model API implementation
    @property
    def universe(self):
        return self._universe

    @property
    def indices(self):
        return self._indices


class AtomSelection(Selection):

    def _max_index(self):
        return self._universe.number_of_atoms

    # Data Model API implementation
    @property
    def type(self):
        return 'atom'


class SiteSelection(Selection):

    def _max_index(self):
        return self._universe.number_of_sites

    # Data Model API implementation
    @property
    def type(self):
        return 'site'


class TemplateAtomSelection(Selection):

    def _max_index(self):
        return self._universe.number_of_template_atoms

    # Data Model API implementation
    @property
    def type(self):
        return 'template_atom'


class TemplateSiteSelection(Selection):

    def _max_index(self):
        return self._universe.number_of_template_sites

    # Data Model API implementation
    @property
    def type(self):
        return 'template_site'


class Factory(AbstractFactory):

    handler = MethodRegister()

    @handler(api.MosaicUniverse)
    def _make_universe(self, mosaic_universe):

        def make_atom(a):
            cons = {"dummy": dummy,
                    "unknown": unknown,
                    "element": element}[a.type]
            return atom(cons(a.name), a.number_of_sites)

        def make_fragment(f):
            fragments = [(sub.label, make_fragment(sub))
                         for sub in f.fragments]
            atoms = [(a.label, make_atom(a)) for a in f.atoms]
            if f.is_polymer:
                return polymer(f.species, fragments, f.bonds, f.polymer_type)
            else:
                return fragment(f.species, fragments, atoms, f.bonds)

        return universe(mosaic_universe.cell_shape,
                        tuple((make_fragment(f), f.label, n)
                              for f, n in mosaic_universe.molecules),
                        mosaic_universe.symmetry_transformations,
                        mosaic_universe.convention)

    @handler(api.MosaicProperty)
    def _make_property(self, mosaic_property):
        universe = self(mosaic_property.universe)
        klass = {"atom": AtomProperty,
                 "site": SiteProperty,
                 "template_atom": TemplateAtomProperty,
                 "template_site": TemplateSiteProperty}[mosaic_property.type]
        return klass(universe, mosaic_property.name, mosaic_property.units,
                     mosaic_property.data)

    @handler(api.MosaicLabel)
    def _make_label(self, mosaic_label):
        universe = self(mosaic_label.universe)
        klass = {"atom": AtomLabel,
                 "site": SiteLabel,
                 "template_atom": TemplateAtomLabel,
                 "template_site": TemplateSiteLabel}[mosaic_label.type]
        return klass(universe, mosaic_label.name, mosaic_label.strings)

    @handler(api.MosaicConfiguration)
    def _make_configuration(self, mosaic_configuration):
        universe = self(mosaic_configuration.universe)
        return Configuration(universe,
                             mosaic_configuration.positions,
                             mosaic_configuration.cell_parameters)

    @handler(api.MosaicSelection)
    def _make_selection(self, mosaic_selection):
        universe = self(mosaic_selection.universe)
        klass = {"atom": AtomSelection,
                 "site": SiteSelection,
                 "template_atom": TemplateAtomSelection,
                 "template_site": TemplateSiteSelection}[mosaic_selection.type]
        return klass(universe, mosaic_selection.indices)



# -*- coding: utf-8 -*-

"""PDB structure import

This module converts structure files from the PDB in mmCIF format to
Mosaic data. The information about the chemical structure of the
residues is taken from the PDB Chemical Component Dictionary.

The code can be considered alpha quality, with standard proteins and
nucleic acid chains being converted correctly. The main difficulty
with handling the unusual cases is ensuring the correct interpretation
of the PDB data.

.. moduleauthor:: Konrad Hinsen <konrad.hinsen@cnrs-orleans.fr>

"""

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

from collections import OrderedDict
import copy
import itertools as it
import os
import sys
import traceback

import numpy as np

from mosaic_pdb.mmcif import MMCIFParser

from mosaic_pdb.space_groups import space_groups

from mosaic_pdb.pdb_chem_comp import components, variants
from mosaic.mutable_model import \
        Element, Atom, Fragment, Polymer, Universe, \
        Configuration, SiteProperty
from mosaic import api

from mosaic.utility import isstring

# Unit conversion factors

class Units:
    Ang = 0.1
    deg = np.pi / 180.


# Crystallographic unit cell

class UnitCell(object):

    def __init__(self, a, b, c, alpha, beta, gamma):
        self.a = a
        self.b = b
        self.c = c
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma

        e1 = np.array([self.a, 0, 0])
        e2 = self.b * np.array([np.cos(self.gamma), np.sin(self.gamma), 0.])
        e3_x = np.cos(self.beta)
        e3_y = (np.cos(self.alpha) - np.cos(self.beta) * np.cos(self.gamma)) \
               / np.sin(self.gamma)
        e3_z = np.sqrt(1. - e3_x ** 2 - e3_y ** 2)
        e3 = self.c * np.array([e3_x, e3_y, e3_z])
        self.basis = (e1, e2, e3)


# An MMCIFStructure accumulates all the information extracted
# from an mmCIF file.

class MMCIFStructure(object):

    def __init__(self, structure_file=None, pdb_code=None):
        """
        Specify the data to be loaded. The following combinations
        are valid:

         - pdb_code only: the data is taken from a public or local
           PDB repository

         - structure_file only: the data is taken from the structure
           file

        @param structure_file: the name of the structure mmCIF file
        @type structure_file: C{str}
        @param pdb_code: a four-letter PDB code
        @type pdb_code: C{str}
        """
        if pdb_code is not None:
            self.pdb_code = pdb_code
            assert structure_file is None
            self.structure_file = None
            parser = MMCIFParser(pdb_code=self.pdb_code)
        elif structure_file is not None:
            self.pdb_code = None
            if isstring(structure_file):
                structure_file = open(structure_file)
            self.structure_file = structure_file
            parser = MMCIFParser(file_object=self.structure_file)
        else:
            raise MMCIFError("No input structure given")

        self.experiment = {}
        self.cell_parameters = {}
        self.symmetry = {}
        self.entities = OrderedDict()
        self.components = OrderedDict()
        self.molecules = OrderedDict()
        self.ss_bridges = []
        self.extra_bonds = []
        self.models = []
        self.u = {}
        self.modified = False
        parser.parseToObjects(cell=self.cell_parameters,
                              symmetry=self.symmetry,
                              exptl=self.experiment,
                              entity=self.addEntity,
                              entity_name_com=self.addEntityName,
                              entity_poly=self.addPolymerInfo,
                              entity_poly_seq=self.buildSequence,
                              chem_comp=self.addComponent,
                              struct_conn=self.addConnection,
                              struct_asym=self.addMolecule,
                              atom_site=self.addSite,
                              atom_sites_alt=self.recordAltLabel,
                              atom_site_anisotrop=self.addAnisoU)
        if self.structure_file is not None:
            self.structure_file.close()
        if 'entry_id' in self.experiment:
            if self.pdb_code is None:
                self.pdb_code = self.experiment['entry_id']
            else:
                assert self.pdb_code == self.experiment['entry_id']
        if len(self.cell_parameters) == 0 \
               or (self.cell_parameters['length_a'] ==
                   self.cell_parameters['length_b'] ==
                   self.cell_parameters['length_a'] == '1'
                   and self.cell_parameters['angle_alpha'] ==
                       self.cell_parameters['angle_beta'] ==
                       self.cell_parameters['angle_gamma'] == '90'):
            # NMR or other non-crystal structure
            self.cell = None
        else:
            self.cell = \
                UnitCell(
                    float(self.cell_parameters['length_a']) * Units.Ang,
                    float(self.cell_parameters['length_b']) * Units.Ang,
                    float(self.cell_parameters['length_c']) * Units.Ang,
                    float(self.cell_parameters['angle_alpha']) * Units.deg,
                    float(self.cell_parameters['angle_beta']) * Units.deg,
                    float(self.cell_parameters['angle_gamma']) * Units.deg)
        del self.cell_parameters
        self.space_group = None
        sg = self.symmetry.get('Int_Tables_number', None)
        if sg not in [None, '?', '.']:
            self.space_group = space_groups[int(sg)]
        else:
            for field in ['pdbx_full_space_group_name_H-M',
                          'space_group_name_H-M']:
                sg = self.symmetry.get(field, None)
                if sg not in [None, '?', '.']:
                    self.space_group = space_groups[sg]
                    break

    def getField(self, label, indices, data):
        if label not in indices:
            return None
        value = data[indices[label]]
        if value == '?':
            return None
        else:
            return value

    def getYesNoField(self, label, indices, data):
        value = self.getField(label, indices, data)
        if value in ['yes', 'y']:
            return True
        elif value in ['no', 'n']:
            return False
        assert value in [None, '.']
        return None

    def addEntity(self, indices, data):
        data = dict((label, data[index])
                    for label, index in indices.items()
                    if data[index] != '?')
        self.entities[data['id']] = data

    def addEntityName(self, indices, data):
        id = data[indices['entity_id']]
        name = data[indices['name']]
        if name != '?':
            self.entities[id]['common_name'] = name

    def addPolymerInfo(self, indices, data):
        entity_id = data[indices['entity_id']]
        polymer_type = self.getField('type', indices, data)
        if polymer_type is not None:
            self.entities[entity_id]['polymer_type'] = polymer_type
            self.entities[entity_id]['polymer_nstd_linkage'] = \
                         self.getYesNoField('nstd_linkage', indices, data)

    def buildSequence(self, indices, data):
        entity_id = data[indices['entity_id']]
        seq = self.entities[entity_id].get('sequence', [])
        self.entities[entity_id]['sequence'] = seq
        res_number = int(data[indices['num']])
        res_type = data[indices['mon_id']]
        if seq and res_number - seq[-1][0] != 1:
            raise ValueError("non-consecutive residue numbers "
                             "in entity_poly_seq")
        seq.append((res_number, res_type))

    def addMolecule(self, indices, data):
        data = dict((label, data[index])
                    for label, index in indices.items()
                    if data[index] != '?')
        self.molecules[data['id']] = data

    def addComponent(self, indices, data):
        data = dict((label, data[index])
                    for label, index in indices.items()
                    if data[index] != '?')
        self.components[data['id']] = data

    def addConnection(self, indices, data):
        if self.getField('conn_type_id', indices, data) == 'disulf':
            asym1, comp1, seq1, asym2, comp2, seq2 = \
                   [self.getField(f, indices, data)
                    for f in ['ptnr1_label_asym_id', 'ptnr1_label_comp_id',
                              'ptnr1_label_seq_id', 'ptnr2_label_asym_id',
                              'ptnr2_label_comp_id', 'ptnr2_label_seq_id']]
            self.ss_bridges.append(((asym1, comp1, int(seq1)),
                                    (asym2, comp2, int(seq2))))
        elif self.getField('conn_type_id', indices, data) == 'covale':
            asym1, comp1, seq1, atom1, asym2, comp2, seq2, atom2 = \
                   [self.getField(f, indices, data)
                    for f in ['ptnr1_label_asym_id', 'ptnr1_label_comp_id',
                              'ptnr1_label_seq_id', 'ptnr1_label_atom_id',
                              'ptnr2_label_asym_id', 'ptnr2_label_comp_id',
                              'ptnr2_label_seq_id', 'ptnr2_label_atom_id']]
            self.extra_bonds.append(((asym1, comp1, seq1, atom1),
                                     (asym2, comp2, seq2, atom2)))

    def recordAltLabel(self, indices, data):
        data = dict((label, data[index])
                    for label, index in indices.items()
                    if data[index] != '?')
        # this information is in principle required for any
        # entry with alternate locations, but in practice it is
        # very rare. This method exists solely to spot such
        # information in order to see if it can be exploited.
        raise ValueError("encountered atom_sites_alt")

    def addSite(self, indices, data):
        asym_id = data[indices['label_asym_id']]
        unique_id = data[indices['id']]
        atom_id = data[indices['label_atom_id']]
        alt_id = data[indices['label_alt_id']]
        comp_id = data[indices['label_comp_id']]

        entity = self.entities[self.molecules[asym_id]['entity_id']]
        is_polymer = entity['type']  == 'polymer' \
                     and 'polymer_type' in entity
        if is_polymer:
            seq_id_string = data[indices['label_seq_id']] 
            try:
                seq_id = int(data[indices['label_seq_id']])
            except ValueError:
                raise ValueError('non-numeric seq_id %s for atom id %s'
                                 % (data[indices['label_seq_id']], unique_id))
            if seq_id <= 0:
                raise ValueError('invalid seq_id %d for atom id %s'
                                 % (seq_id, unique_id))
            ins_code = self.getField('pdbx_PDB_ins_code', indices, data)
        else:
            seq_id = None
            ins_code = None

        try:
            model_number = int(data[indices['pdbx_PDB_model_num']])
        except (KeyError, ValueError):
            model_number = 1
        while len(self.models) < model_number:
            self.models.append(OrderedDict())
        model = self.models[model_number - 1]
        if is_polymer:
            molecule = model.get(asym_id, [[]])
            monomer = molecule[-1]
            if len(monomer) > 0 and \
                   (monomer[0][4] != seq_id or monomer[0][5] != ins_code):
                monomer = []
                molecule.append(monomer)
        else:
            molecule = model.get(asym_id, [])
            monomer = molecule
        model[asym_id] = molecule

        atom_type = data[indices['type_symbol']]
        x = float(data[indices['Cartn_x']]) * Units.Ang
        y = float(data[indices['Cartn_y']]) * Units.Ang
        z = float(data[indices['Cartn_z']]) * Units.Ang
        occupancy = float(data[indices['occupancy']])
        b = self.getField('B_iso_or_equiv', indices, data)
        if b is None:
            u_iso = None
        else:
            u_iso = float(b) * Units.Ang ** 2 / (8. * np.pi ** 2)
        monomer.append((unique_id, atom_id, alt_id, comp_id, seq_id, ins_code,
                        atom_type, x, y, z, occupancy, u_iso))

        stop = False
        for field in ['adp_type', 'aniso_B[1][1]', 'aniso_U[1][1]']:
            value = self.getField(field, indices, data)
            if value is not None:
                sys.stderr.write("%s = %s\n" % (field, str(value)))
                stop = True
        if stop:
            raise ValueError("found aniso data in atom_site")

    def addAnisoU(self, indices, data):
        unique_id = data[indices['id']]
        factor = 1.
        letter = 'U'
        u11 = self.getField('%s[1][1]' % letter, indices, data)
        if u11 is None:
            factor = 1. / (8. * np.pi ** 2)
            letter = 'B'
            if self.getField('%s[1][1]' % letter, indices, data) is None:
                raise ValueError("atom_site_anisotrop with neither U nor B")
        u = np.array([self.getField(letter + ij, indices, data)
                         for ij in ['[1][1]', '[2][2]', '[3][3]',
                                    '[2][3]', '[1][3]', '[1][2]']],
                        dtype=np.float32) * factor * Units.Ang ** 2
        self.u[unique_id] = u


# Make a fragment for a PDB residue/monomer, using the
# chemical component dictionary.

def partition_sites(sites):
    group = []
    last_group_id = None
    for s in sites:
        unique_id, atom_id, alt_id, comp_id, seq_id, ins_code, \
            atom_type, x, y, z, occupancy, u_iso = s
        group_id = (comp_id, seq_id)
        if last_group_id is not None and last_group_id != group_id:
            yield comp_id, group
            group = [s]
        else:
            group.append(s)
        last_group_id = group_id
    if group:
        yield comp_id, group

def make_monomer(sites, site_properties):
    comp_ids = [s[3] for s in sites]
    if len(set(comp_ids)) > 1:
        raise ValueError("Can't handle monomer made of multiple components: %s"
                         % str(set(comp_ids)))
        if False:
            fragments = []
            unique_ids = []
            comp_ids = []
            label_parts = []
            for i, (comp_id, group) in enumerate(partition_sites(sites)):
                f, u = make_monomer(group, site_properties)
                comp_ids.append(comp_id)
                label_parts.append(comp_id + '_' + str(i+1))
                fragments.append(f)
                unique_ids.extend(u)
            bonds = ()  # !!! TODO
            return Fragment('+'.join(label_parts), '+'.join(comp_ids),
                            fragments, (), bonds), \
                   unique_ids       
    atom_ids = set(s[1] for s in sites)
    comp = components[comp_ids[0]]
    comp_atom_ids = set(comp.atoms['atom_id'])
    if hasattr(comp, 'bonds'):
        bond_orders = {'sing': 'single',
                       'doub': 'double',
                       'trip': 'triple',
                       'quad': 'qudruple',
                       'arom': 'aromatic'}
        bonds = tuple((x['atom_id_1'], x['atom_id_2'],
                       bond_orders.get(x['value_order'].lower(), ''))
                      for x in comp.bonds
                      if x['atom_id_1'] in atom_ids
                         and x['atom_id_2'] in atom_ids)
    else:
        bonds = ()
    if not comp_atom_ids.issuperset(atom_ids):
        if atom_ids.difference(comp_atom_ids) == set(["HO5'"]) \
           and "O5'" in atom_ids:
            # Provide a quick fix for a common case in the PDB, to avoid
            # having lots of DNA structures fail for a trivial reason
            comp_atom_ids.add("HO5'")
            bonds = bonds + (("HO5'", "O5'", 'single'),)
        else:
            # Variants exist only for amino acids. Find those that match
            # the atom_ids in the current monomer.
            vs = [v for v in variants.get(comp_ids[0], [])
                  if set(components[v].atoms['atom_id']).issuperset(atom_ids)]
            if not vs:
                missing = atom_ids.difference(comp_atom_ids)
                raise ValueError("Atom(s) %s not in component %s"
                                 % (', '.join(missing), comp_ids[0]))
            variant = None
            if len(vs) == 1:
                variant = vs[0]
            else:
                diff = [len(set(components[v].atoms['atom_id'])
                            .difference(atom_ids))
                        for v in vs]
                if sum(d == 0 for d in diff) == 1:
                    variant = vs[diff.index(0)]
                else:
                    seq_id = sites[0][4]
                    if seq_id == 1:
                        # N-terminal variants have LSN3 in their name
                        vs = [v for v in vs if 'LSN3' in v]
                        if len(vs) == 1:
                            variant = vs[0]
                        else:
                            diff = [len(set(components[v].atoms['atom_id'])
                                        .difference(atom_ids))
                                    for v in vs]
                            if sum(d == 0 for d in diff) == 1:
                                variant = vs[diff.index(0)]
            if variant is None:
                raise ValueError("Monomer variant not unique: %s" % str(vs))
            comp = components[variant]
            comp_atom_ids = set(comp.atoms['atom_id'])
    atom_sites = {}
    atom_ids = []
    atom_types = []
    for unique_id, atom_id, alt_id, comp_id, seq_id, ins_code, \
            atom_type, x, y, z, occupancy, u_iso in sites:
        if atom_id not in comp_atom_ids:
            raise ValueError("No atom named %s in component %s"
                             % (atom_id, comp.id))
        if unique_id in site_properties:
            raise ValueError("Atom id %s not unique" % unique_id)
        if atom_id not in atom_ids:
            atom_ids.append(atom_id)
            atom_types.append(atom_type)
        atom_sites[atom_id] = atom_sites.get(atom_id, [])
        atom_sites[atom_id].append(unique_id)
        site_properties[unique_id] = (x, y, z, occupancy, u_iso)
    atoms = tuple(Atom(atom_id, Element(atom_type.capitalize()),
                       len(atom_sites[atom_id]))
                  for atom_id, atom_type in zip(atom_ids, atom_types))
    unique_ids = tuple(sum((atom_sites[atom_id] for atom_id in atom_ids), []))
    if seq_id is None:
        label = comp.three_letter_code
    else:
        label = comp.three_letter_code + '_' + str(seq_id)
        #if ins_code is not None:
        #    label = label + '_' + ins_code
    return Fragment(label, comp.id, (), atoms, bonds), unique_ids


def ss_bridges(structure, asym_id_1, asym_id_2):
    """
    Iterate over the disulfide bridges in the structure.
    """
    for (asym1, comp1, seq1), (asym2, comp2, seq2) \
            in structure.ss_bridges:
        if asym1 == asym_id_1 and asym2 == asym_id_2:
            yield seq1, seq2
        elif asym1 == asym_id_2 and asym2 == asym_id_1:
            yield seq2, seq1


def distinct_names(fragments):
    """
    Make sure that all the fragments have distinct names,
    adding distinctive integer suffixes if necessary.
    """
    name_counts = {}
    for f in fragments:
        name_counts[f.label] = name_counts.get(f.label, 0) + 1
    new_names = {}
    for name, count in name_counts.items():
        if count == 1:
            new_names[name] = [name]
        else:
            new_names[name] = [name + '_' + str(n)
                               for n in range(1, count + 1)]
    for f in fragments:
        label = new_names[f.label][0]
        del new_names[f.label][0]
        f.label = label


class MoleculeList(list):

    """
    Turn a sequence of molecules into a compressed molecule list
    with repetition counts.
    """

    def __init__(self, iterable):
        list.__init__(self, [])
        for item in iterable:
            self.append(item)

    def append(self, item):
        assert isinstance(item, Fragment)
        if len(self) == 0:
            list.append(self, (item, 1))
        else:
            last, count = self[-1]
            if last.eq_structure(item):
                if last.label == item.label:
                    f = last
                else:
                    f = copy.copy(last)
                    f.label = last.label + ',' + item.label
                self[-1] = (f, count + 1)
            else:
                list.append(self, (item, 1))


def symmetry_transformations(space_group):
    """
    Convert the symmetry transformations for the crystal's space
    group to the right format.
    """
    transformations = []
    assert (space_group.transformations[0][0] ==
            np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])).all()
    assert (space_group.transformations[0][1] == np.array([0, 0, 0])).all()
    assert (space_group.transformations[0][2] == np.array([1, 1, 1])).all()
    for rot, tr_num, tr_den in space_group.transformations[1:]:
        rot = np.array(rot, dtype=np.float64)
        tr = np.array(tr_num, dtype=np.float64) / \
             np.array(tr_den, dtype=np.float64)
        transformations.append((rot, tr))
    return transformations


def make_crystal(structure, model_number, universes):
    """
    Make a Mosaic representation of a single model for a crystal.
    Add the universe to the passed-in universe list if it's not
    already in it. This takes care of the common situation that
    multiple models share a common universe.
    """
    model = structure.models[model_number]
    site_properties = {}
    monomer_fragments = {}
    molecules = {}
    unique_ids = {}
    for asym_id, sites in model.items():
        unique_ids[asym_id] = []
        entity_id = structure.molecules[asym_id]['entity_id']
        entity = structure.entities[entity_id]
        entity_type = entity['type']
        if entity_type == 'polymer' and 'polymer_type' in entity:
            # Sugars are classified as polymers in the PDB, but have
            # no polymer type and a single residue with the
            # non-numeric seq_id '.'. This looks like a bug in the PDB
            # data, but since it's so frequent, we have to deal with
            # it. We do so by treating polymers without a polymer_type
            # as non-polymers.
            polymer_type = entity['polymer_type']
            if entity['polymer_nstd_linkage']:
                raise ValueError("Non-standard linkage not yet implemented")
            if 'sequence' not in entity:
                raise ValueError('Polymer entity has no sequence information')
            name = entity.get('pdbx_description', None)
            if name is None:
                name = entity.get('common_name', None)
            if name is None:
                name = "Entity_" + entity_id
            name = name.replace(' ', '_')
            name = name.replace('.', '_')
            # Start with empty residues
            residues = OrderedDict()
            for seq_id, comp_id in entity['sequence']:
                f = Fragment(comp_id + '_' + str(seq_id), comp_id,
                             (), (), ())
                residues[seq_id] = f
                monomer_fragments[(asym_id, seq_id)] = f
            # For residues that have sites, make non-empty fragments
            for monomer_sites in sites:
                assert structure.modified or len(monomer_sites) > 0
                if not monomer_sites:
                    continue
                seq_id = monomer_sites[0][4]
                f, u = make_monomer(monomer_sites, site_properties)
                species = monomer_fragments[(asym_id, seq_id)].species
                assert f.species.startswith(species)
                monomer_fragments[(asym_id, seq_id)] = f
                if seq_id not in residues:
                    raise ValueError('seq_id %d in site list but not '
                                     'in sequence' % seq_id)
                residues[seq_id] = f
                unique_ids[asym_id].extend(u)
            residues = list(residues.values())
            # Add inter-monomer links
            if polymer_type in ['polypeptide(L)']:
                peptide_bonds = tuple((r1.get('C'), r2.get('N'), 'single')
                                      for r1, r2 in zip(residues[:-1],
                                                        residues[1:]))
                ss_bonds = tuple((monomer_fragments[(asym_id, seq1)].get('SG'),
                                  monomer_fragments[(asym_id, seq2)].get('SG'),
                                  'single')
                                 for seq1, seq2 in
                                     ss_bridges(structure, asym_id, asym_id))
                bonds = peptide_bonds + ss_bonds
                polymer_type = 'polypeptide'
            elif polymer_type in ['polyribonucleotide',
                                  'polydeoxyribonucleotide',
                                  'polydeoxyribonucleotide/'
                                      'polyribonucleotide hybrid']:
                bonds = tuple((r1.get('O3\''), r2.get('P'), 'single')
                              for r1, r2 in zip(residues[:-1], residues[1:]))
            elif polymer_type in ['polysaccharide(D)', 'polysaccharide(L)']:
                # need to figure out how monosaccharides are linked
                # example: 1AGA
                bonds = tuple()
            else:
                raise ValueError("Polymer type %s not yet implemented"
                                 % polymer_type)
            bonds = tuple((a1, a2, order)
                          for a1, a2, order in bonds
                          if a1 is not None and a2 is not None)
            polymer_types = {'polydeoxyribonucleotide/'
                                 'polyribonucleotide hybrid':
                             'polynucleotide'}
            # Cleanup name to make it a valid species name
            name = ''.join(c for c in name if c in api._allowed_in_labels)
            if len(name) > 32767:
                name = name[:32767]
            molecules[asym_id] = [Polymer(asym_id, name, residues, bonds,
                                          polymer_types.get(polymer_type,
                                                            polymer_type))]
        elif entity_type == 'water':
            ms = []
            for atom in sites:
                if atom[6] == 'O':
                    ms.append([atom])
                else:
                    assert atom[6] == 'H'
                    ms[-1].append(atom)
            molecules[asym_id] = []
            for m in ms:
                f, u = make_monomer(m, site_properties)
                molecules[asym_id].append(f)
                unique_ids[asym_id].extend(u)
        else:
            f, u = make_monomer(sites, site_properties)
            molecules[asym_id] = [f]
            unique_ids[asym_id].extend(u)

    # Treat inter-chain disulfide bridges
    order_kept = not structure.modified
    clusters = []
    mol_to_cluster = {}
    for a in model.keys():
        mol_to_cluster[a] = len(clusters)
        clusters.append((set([a]), []))
    for asym1, asym2 in it.combinations(model.keys(), 2):
        ss = list(ss_bridges(structure, asym1, asym2))
        if ss:
            c1 = mol_to_cluster[asym1]
            c2 = mol_to_cluster[asym2]
            if c1 != c2:
                clusters[c1][0].update(clusters[c2][0])
                clusters[c1][1].extend(clusters[c2][1])
                clusters[c2] = (None, None)
                mol_to_cluster[asym2] = c1
            clusters[c1][1].extend([(asym1, s1, asym2, s2) for s1, s2 in ss])
    for asym_ids, bridges in clusters:
        if asym_ids is not None and len(asym_ids) > 1:
            asym_ids = list(asym_ids)
            order_kept = False
            # This could be improved to maintain the atom order in
            # specific and frequent special cases. For later...
            fragments = sum((molecules[a] for a in asym_ids), [])
            distinct_names(fragments)
            uids = sum((unique_ids[a] for a in asym_ids), [])
            bonds = tuple((monomer_fragments[(asym1, seq1)].get('SG'),
                           monomer_fragments[(asym2, seq2)].get('SG'),
                           'single')
                          for asym1, seq1, asym2, seq2 in bridges)
            bonds = tuple((a1, a2, order)
                          for a1, a2, order in bonds
                          if a1 is not None and a2 is not None)
            mol = Fragment('+'.join(f.label for f in fragments),
                           '+'.join(f.species for f in fragments),
                           fragments, (), bonds)
            molecules[asym_ids[0]] = [mol]
            unique_ids[asym_ids[0]] = uids
            for a in asym_ids[1:]:
                del molecules[a]
                del unique_ids[a]

    if structure.cell is None:
        cell_shape = 'infinite'
        st = []
    else:
        pdb_cell_shape = structure.symmetry.get('cell_setting', None)
        cell_shape = {'cubic': 'cube',
                      'orthorhombic': 'cuboid',
                      'tetragonal': 'cuboid',
                      }.get(pdb_cell_shape, 'parallelepiped')
        if structure.space_group is None:
            st = []
        else:
            st = symmetry_transformations(structure.space_group)
    universe = Universe(cell_shape,
                        MoleculeList(m
                                     for m in it.chain.from_iterable(
                                         molecules.get(asym_id, [])
                                         for asym_id in model.keys())),
                        symmetry_transformations=st,
                        convention='PDB')
    assert universe.number_of_sites == len(site_properties)

    # Check if an equal universe already exists
    if universe in universes:
        universe = universes[universes.index(universe)]
    else:
        universes.append(universe)

    # Configuration and occupancy
    positions = np.zeros((universe.number_of_sites, 3), np.float32)
    occupancy = np.zeros((universe.number_of_sites,), np.float32)
    u_iso = np.zeros((universe.number_of_sites,), np.float32)
    for i, uid in enumerate(sum((unique_ids.get(asym_id, [])
                                 for asym_id in model.keys()), [])):
        # The conversion process should maintain the order of the sites
        # in the mmCIF file, as long as multiple sites corresponding to the
        # same atom (different alt_ids) occur consecutively in the list.
        # This is not explicitly promised by the documentation, but seems
        # to be the case in most if not all real-life mmCIF files.
        # Some specific features (e.g. disulfide bonds between chains)
        # can cause the site order to be changed.
        if order_kept and model_number == 0:
            assert i == int(uid) - 1
        positions[i, :] = site_properties[uid][0:3]
        occupancy[i] = site_properties[uid][3]
        if u_iso is not None and site_properties[uid][4] is not None:
            u_iso[i] = site_properties[uid][4]
    if universe.cell_shape == 'infinite':
        cell_parameters = None
    else:
        cell_parameters = np.zeros((3, 3), np.float32)
        for i, v in enumerate(structure.cell.basis):
            cell_parameters[i, :] = v
        if universe.cell_shape == 'cube':
            edge = cell_parameters[0, 0]
            assert np.minimum.reduce(
                        np.fabs((cell_parameters
                                    - edge * np.eye(3)).flat)) < 1.e-7
            cell_parameters = edge
        elif universe.cell_shape == 'cuboid':
            edge1 = cell_parameters[0, 0]
            edge2 = cell_parameters[1, 1]
            edge3 = cell_parameters[2, 2]
            assert np.minimum.reduce(
                np.fabs((cell_parameters -
                            np.array([[edge1, 0., 0.],
                                         [0., edge2, 0.],
                                         [0., 0., edge3]])).flat)) < 1.e-7
            cell_parameters = np.array([edge1, edge2, edge3],
                                          dtype=np.float32)

    data = OrderedDict()
    data['configuration'] = Configuration(universe, positions, cell_parameters)
    if not (occupancy == 1.).all():
        data['occupancy'] = SiteProperty(universe, "occupancy", "", occupancy)

    # Anisotropic or isotropic U
    if len(structure.models) == 1 and len(structure.u) > 0:
        u_aniso = np.zeros((universe.number_of_sites, 6), np.float32)
        for i, uid in enumerate(sum((unique_ids.get(asym_id, [])
                                     for asym_id in model.keys()), [])):
            try:
                u_aniso[i] = structure.u[uid]
            except KeyError:
                u_aniso[i, :3] = u_iso[i]
                u_aniso[i, 3:] = 0.
        data['anisotropic_displacement_parameter'] = \
            SiteProperty(universe, 'anisotropic_displacement_parameter',
                         'nm2', u_aniso)
    elif not (u_iso == 0.).all():
        data['isotropic_displacement_parameter'] = \
            SiteProperty(universe, 'isotropic_displacement_parameter',
                         'nm2', u_iso)

    return data


def make_models(structure_file=None, pdb_code=None):
    """
    Return an iterator over all the data items needed to
    represent the structure file's contents. The first
    items are the universes, then everything else.
    """
    s = MMCIFStructure(structure_file=structure_file,
                       pdb_code=pdb_code)
    universes = []
    models = [make_crystal(s, i, universes)
              for i in range(len(s.models))]

    if len(universes) == 1:
        yield "universe", universes[0]
    else:
        for i in range(len(universes)):
            yield "universe_%d" % (i + 1), universes[i]

    if len(models) == 1:
        for name, value in models[0].items():
            yield name, value
    else:
        for i, model in enumerate(models):
            for name, value in model.items():
                yield "%s_model_%d" % (name, i + 1), value



# -*- coding: utf-8 -*-
"""XML I/O

.. moduleauthor:: Konrad Hinsen

The XML representation of Mosaic data consists of a single element
``<mosaic version="x.y">...</mosaic>`` that can contain any number of
Mosaic data items, each of which is identified by a unique id.
The classes :class:`XMLWriter` and :class:`XMLReader` handle the
translation between Mosaic data items in memory that can reference
each other and XML elements that reference each other by id.

A common pattern for generating an XML file is:

::

    with XMLWriter('molecule.xml') as writer:
       writer.store("universe", universe)
       writer.store("configuration1", configuration1)
       writer.store("configuration2", configuration2)

A common pattern for reading from an XML file is:

::

    items = {}
    for id, data in XMLReader('molecule.xml'):
       items[id] = data

"""

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

try:
    import xml.etree.cElementTree as ET
except IOError:
    import xml.etree.ElementTree as ET

import numpy as N
import immutable.np as IN

import mosaic.api as api
import mosaic.immutable_model as im
from mosaic.utility import MethodRegister
from mosaic.utility import uint_for_max_value
from mosaic.utility import isstring
from mosaic.utility import xml_encoding


# Number formatting

def n2s(x, dp=False):
    if isinstance(x, int):
        return str(x)
    else:
        if dp:
            return "{0:.20g}".format(x)
        else:
            return "{0:.10g}".format(x)


# Data type formatting

def t2s(t):
    return {N.dtype(N.int8): "int8",
            N.dtype(N.int16): "int16",
            N.dtype(N.int32): "int32",
            N.dtype(N.int32): "int32",
            N.dtype(N.uint8): "uint8",
            N.dtype(N.uint16): "uint16",
            N.dtype(N.uint32): "uint32",
            N.dtype(N.uint32): "uint32",
            N.dtype(N.float32): "float32",
            N.dtype(N.float64): "float64",
            N.dtype(N.bool): "boolean"
            }[t]


def s2t(s):
    return {"int8": N.int8,
            "int16": N.int16,
            "int32": N.int32,
            "int32": N.int32,
            "uint8": N.uint8,
            "uint16": N.uint16,
            "uint32": N.uint32,
            "uint32": N.uint32,
            "float32": N.float32,
            "float64": N.float64,
            "boolean": N.bool
            }[s]

class XMLStore(object):

    def __init__(self):
        self._id_map = {}
        self._data_map = {}

    def _register_data_item(self, xml_id, data_item):
        self._data_map[xml_id] = data_item
        self._id_map[data_item] = xml_id

    def _get_id(self, data_item):
        return self._id_map.get(data_item, None)

    def _get_data(self, xml_id):
        return self._data_map.get(xml_id, None)

    # Make XMLStores work as context managers
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False


class XMLWriter(XMLStore):

    """Output handler for XML files

    This class handles the translation of references between Mosaic
    data items in memory to XML references by id. References are
    allowed only to data items that have been written earlier using
    the same XMLWriter instance.

    An XMLWriter can be used like a file, in which case it must be
    closed after the last ``store`` operation, or as a context manager
    in a with-statement.
    """

    def __init__(self, xml_file):
        """
        :param xml_file: a writeable file object, or a string
                         interpreted as a file name
        :type xml_file: str or file-like
        """
        XMLStore.__init__(self)
        if isstring(xml_file):
            self.file = open(xml_file, 'w')
            self._close_file = True
        else:
            # assume it is a file-like object
            self.file = xml_file
            self._close_file = False
        # Keep a set of already used ids in order to check they remain unique
        self._xml_ids = set()
        # Start with the XML declaration
        self.file.write('<?xml version="1.0" encoding="utf-8"?>')
        # Open the top-level element
        self.file.write('<mosaic version="%d.%d">' % api.MOSAIC_VERSION)

    def close(self):
        # Close the top-level element
        self.file.write("</mosaic>")
        # Close the underlying file if we own it
        if self._close_file:
            self.file.close()
        # No more output after this
        self.file = None
        # Clear cache
        self._id_map = None
        self._data_map = None

    storage_handler = MethodRegister()

    def store(self, xml_id, data):
        """
        :param xml_id: the id of the XML element representing the data item
        :type xml_id:  str
        :param data:   a Mosaic data item
        :type data:    :class:`mosaic.api.MosaicDataItem`
        """
        api.validate_type(xml_id, str, "xml_id")
        api.validate_type(data, api.MosaicDataItem, "data")
        xml_id = xml_id
        if xml_id in self._xml_ids:
            raise ValueError("XML ID %s has already been used" % xml_id)
        handler = self.storage_handler[type(data)]
        if handler is None:
            raise TypeError("Storage of %s not yet implemented"
                            % str(type(data)))
        handler(self, xml_id, data)
        self._register_data_item(xml_id, data)
        self._xml_ids.add(xml_id)

    def _write(self, element):
        ET.ElementTree(element).write(self.file,
                                      encoding=xml_encoding)

    #
    # Storage handlers
    #
    @storage_handler(api.MosaicUniverse)
    def _store_universe(self, xml_id, universe):
        el = ET.Element('universe')
        el.set("id", xml_id)
        el.set("cell_shape", universe.cell_shape)
        el.set("convention", universe.convention)
        st = universe.symmetry_transformations
        if len(st) > 0:
            el_st = ET.SubElement(el, 'symmetry_transformations')
            for rot, trans in st:
                el_st.append(self._symmetry_el(rot, trans))
        el_molecules = ET.SubElement(el, 'molecules')
        for fragment, count in universe.molecules:
            el_m = ET.Element('molecule', count=str(count))
            el_m.append(self._fragment_element(fragment))
            el_molecules.append(el_m)
        self._write(el)

    def _symmetry_el(self, rot, trans):
        el = ET.Element('transformation')
        el_rot = ET.SubElement(el, 'rotation')
        el_rot.text = ' '.join(n2s(x) for x in rot.flat)
        el_trans = ET.SubElement(el, 'translation')
        el_trans.text = ' '.join(n2s(x) for x in trans.flat)
        return el

    def _fragment_element(self, fragment):
        el = ET.Element('fragment',
                        label=fragment.label,
                        species=fragment.species)
        if fragment.is_polymer:
            el.set('polymer_type', fragment.polymer_type)
        if len(fragment.fragments) > 0:
            el_fragments = ET.SubElement(el, 'fragments')
            for f in fragment.fragments:
                el_fragments.append(self._fragment_element(f))
        if len(fragment.atoms) > 0:
            el_atoms = ET.SubElement(el, 'atoms')
            for a in fragment.atoms:
                el_atom = ET.Element('atom',
                                     label=a.label,
                                     type=a.type,
                                     name=a.name)
                if a.number_of_sites != 1:
                    el_atom.set('nsites', str(a.number_of_sites))
                el_atoms.append(el_atom)
        if len(fragment.bonds) > 0:
            el_bonds = ET.SubElement(el, 'bonds')
            for a1, a2, order in fragment.bonds:
                el_bonds.append(ET.Element('bond',
                                           atoms=a1 + ' ' + a2,
                                           order=order))
        return el

    @storage_handler(api.MosaicConfiguration)
    def _store_configuration(self, xml_id, configuration):
        universe = configuration.universe
        universe_id = self._get_id(universe)
        if universe_id is None:
            raise IOError("universe must be stored first")
        el = ET.Element('configuration')
        el.set("id", xml_id)
        el.append(ET.Element('universe', ref=universe_id))
        float_type = t2s(configuration.positions.dtype)
        dp = float_type == "float64"
        if N.product(configuration.cell_parameters.shape) != 0:
            shape_str = ' '.join(str(x)
                                 for x in configuration.cell_parameters.shape)
            el_cp = ET.Element('cell_parameters', shape=shape_str)

            el_cp.text = ' '.join(n2s(p, dp)
                                  for p in configuration.cell_parameters.flat)
            el.append(el_cp)
        el_pos = ET.Element('positions', type=float_type)
        el_pos.text = ' '.join(' '.join(n2s(x, dp) for x in p)
                               for p in configuration.positions)
        el.append(el_pos)
        self._write(el)

    @storage_handler(api.MosaicProperty)
    def _store_property(self, xml_id, property):
        universe = property.universe
        universe_id = self._get_id(universe)
        if universe_id is None:
            raise IOError("universe must be stored first")
        el = ET.Element(property.type + '_property')
        el.set("id", xml_id)
        el.set("name", property.name)
        el.set("units", property.units)
        el.append(ET.Element('universe', ref=universe_id))
        shape_str = ' '.join(str(x) for x in property.element_shape)
        el_data = ET.Element('data',
                             shape=shape_str,
                             type=t2s(property.data.dtype))
        dp = property.data.dtype == N.dtype(N.float64)
        el_data.text = ' '.join(' '.join(n2s(x, dp) for x in v.flat)
                                for v in property.data)
        el.append(el_data)
        self._write(el)

    @storage_handler(api.MosaicLabel)
    def _store_label(self, xml_id, label):
        universe = label.universe
        universe_id = self._get_id(universe)
        if universe_id is None:
            raise IOError("universe must be stored first")
        el = ET.Element(label.type + '_label')
        el.set("id", xml_id)
        el.set("name", label.name)
        el.append(ET.Element('universe', ref=universe_id))
        el_strings = ET.Element("strings")
        el_strings.text = ' '.join(label.strings)
        el.append(el_strings)
        self._write(el)

    @storage_handler(api.MosaicSelection)
    def _store_selection(self, xml_id, selection):
        universe = selection.universe
        universe_id = self._get_id(universe)
        if universe_id is None:
            raise IOError("universe must be stored first")
        el = ET.Element(selection.type + '_selection')
        el.set("id", xml_id)
        el.append(ET.Element('universe', ref=universe_id))
        el_indices = ET.Element("indices")
        el_indices.text = ' '.join(str(i) for i in selection.indices)
        el.append(el_indices)
        self._write(el)


class XMLReader(XMLStore):

    """Input handler for XML files

    This class handles the translation of references by id in an XML
    file to in-memory references between data items.

    An XMLReader is used as an iterator over ``(id, data_item)`` pairs.

    The current implementation does not do any validation, it assumes
    the XML input to be correct.
    """


    def __init__(self, xml_file):
        """
        :param xml_file: a file object, or a string
                         interpreted as a file name
        :type xml_file: str or file-like
        """
        XMLStore.__init__(self)
        if isstring(xml_file):
            # file name given: open file and close it at the end
            self.file = open(xml_file, 'r')
            self._close_file = True
        else:
            # assume it is a file-like object, don't close at the end
            self.file = xml_file
            self._close_file = False

    def close(self):
        if self._close_file:
            self.file.close()
        # No more reading after this
        self.file = None
        self._close_file = False
        # Clear cache
        self._id_map = None
        self._data_map = None

    def __del__(self):
        self.close()

    def __iter__(self):
        for event, el in ET.iterparse(self.file):
            if el.tag == 'mosaic':
                # check version number
                version = tuple(int(s) for s in el.get("version").split('.'))
                if version[0] > api.MOSAIC_VERSION[0] \
                   or version[1] > api.MOSAIC_VERSION[1]:
                    raise ValueError("XML data is for version %d.%d, "
                                     "software is version %d.%d"
                                     % (version + api.MOSAIC_VERSION))
                continue
            xml_id = el.get("id")
            if xml_id is not None:
                handler = self.data_handler[el.tag]
                if handler is None:
                    raise ValueError("Unknown element type %s" % el.tag)
                data = handler(self, el)
                el.clear()
                self._register_data_item(xml_id, data)
                yield xml_id, data

    data_handler = MethodRegister()

    @data_handler("universe")
    def _read_universe(self, el):
        ref = el.get('ref', None)
        if ref is not None:
            return self._get_data(ref)
        molecules = \
          tuple(self._parse_molecule(el_m)
                for el_m in el.iter('molecule'))
        symmetry_transformations = \
          tuple((self._array((3, 3), N.float64, el_t.find('rotation').text),
                 self._array((3,), N.float64, el_t.find('translation').text))
                for el_t in el.iter('transformation'))
        return im.universe(el.get('cell_shape'),
                           molecules,
                           symmetry_transformations,
                           el.get('convention'))

    def _parse_molecule(self, el):
        label, fragment = self._parse_fragment_tree(el.find('fragment'))
        count = int(el.get('count'))
        return fragment, label, count

    def _parse_fragment_tree(self, el):
        fragments = []
        for el_fragments in el.findall('fragments'):
            for el_f in el_fragments:
                fragments.append(self._parse_fragment_tree(el_f))
        atoms = []
        for el_atoms in el.findall('atoms'):
            for el_a in el_atoms:
                atom_classes = {"dummy": im.dummy,
                                "unknown": im.unknown,
                                "element": im.element}
                descr = atom_classes[el_a.get('type')](el_a.get('name'))
                atoms.append((el_a.get('label'),
                              im.atom(descr, int(el_a.get('nsites', '1')))))
        bonds = []
        for el_bonds in el.findall('bonds'):
            for el_b in el_bonds:
                a1, a2 = el_b.get('atoms').split()
                bonds.append((a1, a2, el_b.get('order')))
        label = el.get('label')
        species = el.get('species')
        polymer_type = el.get('polymer_type', None)
        if polymer_type is None:
            return label, im.fragment(species, fragments, atoms, bonds)
        else:
            assert len(atoms) == 0
            return label, im.polymer(species, fragments, bonds, polymer_type)

    @data_handler("configuration")
    def _read_configuration(self, el):
        universe = self._read_universe(el.find('universe'))
        el_pos = el.find('positions')
        dtype = s2t(el_pos.get('type'))
        pos = self._property((3,), dtype, el_pos.text)
        el_cp = el.find('cell_parameters')
        if el_cp is None:
            cp = None
        else:
            cp = self._array(el_cp.get('shape'), dtype, el_cp.text)
        return im.Configuration(universe, pos, cp)

    @data_handler("atom_property")
    @data_handler("site_property")
    @data_handler("template_atom_property")
    @data_handler("template_site_property")
    def _read_property(self, el):
        universe = self._read_universe(el.find('universe'))
        ptype = el.tag[:-9]   # strip off '_property'
        klass = {"atom": im.AtomProperty,
                 "site": im.SiteProperty,
                 "template_atom": im.TemplateAtomProperty,
                 "template_site": im.TemplateSiteProperty}[ptype]
        el_d = el.find('data')
        data = self._property(el_d.get('shape'),
                              s2t(el_d.get('type')),
                              el_d.text)
        return klass(universe, el.get('name'), el.get('units'), data)

    @data_handler("atom_label")
    @data_handler("site_label")
    @data_handler("template_atom_label")
    @data_handler("template_site_label")
    def _read_label(self, el):
        universe = self._read_universe(el.find('universe'))
        ptype = el.tag[:-6]   # strip off '_label'
        klass = {"atom": im.AtomLabel,
                 "site": im.SiteLabel,
                 "template_atom": im.TemplateAtomLabel,
                 "template_site": im.TemplateSiteLabel}[ptype]
        el_strings = el.find('strings')
        strings = tuple(el_strings.text.split())
        return klass(universe, el.get('name'), strings)

    @data_handler("atom_selection")
    @data_handler("site_selection")
    @data_handler("template_atom_selection")
    @data_handler("template_site_selection")
    def _read_selection(self, el):
        universe = self._read_universe(el.find('universe'))
        ptype = el.tag[:-10]   # strip off '_selection'
        klass = {"atom": im.AtomSelection,
                 "site": im.SiteSelection,
                 "template_atom": im.TemplateAtomSelection,
                 "template_site": im.TemplateSiteSelection}[ptype]
        el_d = el.find('indices')
        indices = [int(s) for s in el_d.text.split()]
        indices = IN.array(indices, uint_for_max_value(max(indices)))
        return klass(universe, indices)

    # Parse array data

    def _array(self, shape, dtype, text):
        if isstring(shape):
            shape = tuple(int(s) for s in shape.split())
        return IN.array([dtype(s) for s in text.split()]).reshape(shape)

    def _property(self, el_shape, dtype, text):
        data = IN.array([dtype(s) for s in text.split()])
        if isstring(el_shape):
            el_shape = tuple(int(s) for s in el_shape.split())
        n_els = N.product(el_shape, dtype=N.int)
        assert len(data) % n_els == 0
        n_entries = len(data) // n_els
        return data.reshape((n_entries,)+el_shape)



# -*- coding: utf-8 -*-

# Tests for the PDB importer

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import unittest

from mosaic.xml_io import XMLReader
from mosaic_pdb.import_structure import make_models

class PDB2ONXTest(unittest.TestCase):

    def setUp(self):
        self.pdb_model = dict(make_models("2ONX.cif"))
        self.xml_model = dict(XMLReader("2ONX.xml"))

    def test_equivalence(self):
        for name, data in self.pdb_model.items():
            self.assertTrue(name in self.xml_model)
            self.assertTrue(data.is_equivalent(self.xml_model[name]))

def suite():
    loader = unittest.TestLoader()
    s = unittest.TestSuite()
    s.addTest(loader.loadTestsFromTestCase(PDB2ONXTest))
    return s

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# Tests for module mosaic.xml_io

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import unittest
import copy
import io
import sys

import numpy as N
import numpy.random as NR
import immutable.np as IN

import mosaic.mutable_model
import mosaic.immutable_model
import mosaic.array_model
from mosaic.api import is_valid
from mosaic.xml_io import XMLWriter, XMLReader

if sys.version_info[0] == 2:
    def StringIO(buffer=None):
        return io.BytesIO(buffer)
else:
    def StringIO(buffer=None):
        return io.StringIO(buffer)

class PeptideTest(object):

    def test_universe(self):
        file = StringIO()
        with XMLWriter(file) as writer:
            writer.store("universe", self.universe)
        xml_string = file.getvalue()
        file = StringIO(xml_string)
        with XMLReader(file) as reader:
            data = {xml_id: obj for xml_id, obj in reader}
        universe = data['universe']
        self.assertTrue(universe.is_equivalent(self.universe))
        self.assertTrue(is_valid(universe))

    def test_configuration(self):
        M = self.M
        for dtype in (N.float32, N.float64):
            conf = M.Configuration(self.universe,
                                   IN.array(NR.normal(size=(40,3)),
                                            dtype=dtype),
                                   IN.array(10., dtype))
            self.assertTrue(is_valid(conf))
            file = StringIO()
            with XMLWriter(file) as writer:
                writer.store("universe", self.universe)
                writer.store("configuration", conf)
            xml_string = file.getvalue()
            file = StringIO(xml_string)
            with XMLReader(file) as reader:
                data = {xml_id: obj for xml_id, obj in reader}
            conf_r = data['configuration']
            self.assertTrue(conf_r.is_equivalent(conf))
            self.assertTrue(is_valid(conf_r))

    def test_property(self):
        M = self.M
        universe = self.universe
        for dtype in (N.float32, N.float64):
            for p in [M.AtomProperty(universe, "mass", "amu",
                         IN.array(NR.normal(size=(universe.number_of_atoms,)))),
                      M.SiteProperty(universe, "foo", "",
                         IN.array(NR.normal(size=(universe.number_of_sites, 3)))),
                      M.TemplateAtomProperty(universe, "mass", "amu",
                         IN.array(NR.normal(size=(universe.number_of_template_atoms,
                                                  1)))),
                      M.TemplateSiteProperty(universe, "foo", "bar",
                         IN.array(NR.normal(size=(universe.number_of_template_sites,
                                                  2, 2)))),
                      ]:
                self.assertTrue(is_valid(p))
                file = StringIO()
                with XMLWriter(file) as writer:
                    writer.store("universe", self.universe)
                    writer.store("property", p)
                xml_string = file.getvalue()
                file = StringIO(xml_string)
                with XMLReader(file) as reader:
                    data = {xml_id: obj for xml_id, obj in reader}
                p_r = data['property']
                self.assertTrue(p_r.is_equivalent(p))
                self.assertTrue(is_valid(p_r))

    def test_label(self):
        M = self.M
        universe = self.universe
        for l in [M.AtomLabel(universe, "element",
                              tuple(a.name
                                    for f, n in universe.molecules
                                    for _ in range(n)
                                    for a in f.recursive_atom_iterator())),
                  M.TemplateAtomLabel(universe, "element",
                                      tuple(a.name
                                            for f, n in universe.molecules
                                            for a in f.recursive_atom_iterator())),
                  M.SiteLabel(universe, "element",
                              tuple(a.name
                                    for f, n in self.universe.molecules
                                    for _ in range(n)
                                    for a in f.recursive_atom_iterator()
                                    for __ in range(a.number_of_sites))),
                  M.TemplateSiteLabel(universe, "element",
                                      tuple(a.name
                                            for f, n in self.universe.molecules
                                            for a in f.recursive_atom_iterator()
                                            for _ in range(a.number_of_sites))),
                  ]:
            self.assertTrue(is_valid(l))
            file = StringIO()
            with XMLWriter(file) as writer:
                writer.store("universe", self.universe)
                writer.store("label", l)
            xml_string = file.getvalue()
            file = StringIO(xml_string)
            with XMLReader(file) as reader:
                data = {xml_id: obj for xml_id, obj in reader}
            l_r = data['label']
            self.assertTrue(l_r.is_equivalent(l))
            self.assertTrue(is_valid(l_r))

    def test_selection(self):
        M = self.M
        universe = self.universe
        for s in [M.AtomSelection(universe, IN.array([0], N.uint8)),
                  M.SiteSelection(universe, IN.array([1, 2], N.uint8)),
                  M.TemplateAtomSelection(universe, IN.array([0, 2], N.uint8)),
                  M.TemplateSiteSelection(universe, IN.array([0, 3], N.uint8))]:
            self.assertTrue(is_valid(s))
            file = StringIO()
            with XMLWriter(file) as writer:
                writer.store("universe", self.universe)
                writer.store("selection", s)
            xml_string = file.getvalue()
            file = StringIO(xml_string)
            with XMLReader(file) as reader:
                data = {xml_id: obj for xml_id, obj in reader}
            s_r = data['selection']
            self.assertTrue(s_r.is_equivalent(s))
            self.assertTrue(is_valid(s_r))


class MMPeptideTest(unittest.TestCase,
                    PeptideTest):
    M = mosaic.mutable_model

    def setUp(self):
        M = self.M
        C = M.Element('C')
        H = M.Element('H')
        N = M.Element('N')
        O = M.Element('O')
        peptide_group = M.Fragment('peptide', 'peptide_group',
                                   (),
                                   (M.Atom('CA', C),
                                    M.Atom('HA', H),
                                    M.Atom('H', H),
                                    M.Atom('N', N),
                                    M.Atom('C', C),
                                    M.Atom('O', O)),
                                    (('N', 'H', 'single'),
                                     ('N', 'CA', 'single'),
                                     ('CA', 'HA', 'single'),
                                     ('CA', 'C', 'single'),
                                     ('C', 'O', 'double')))
        ala_sidechain = M.Fragment('sidechain', 'ala_sidechain',
                                   (),
                                   (M.Atom('CB', C),
                                    M.Atom('HB1', H),
                                    M.Atom('HB2', H),
                                    M.Atom('HB3', H)),
                                   (('CB', 'HB1', 'single'),
                                    ('CB', 'HB2', 'single'),
                                    ('CB', 'HB3', 'single'),))
        ala = lambda label: M.Fragment(label, 'alanine',
                                       (copy.copy(peptide_group),
                                        copy.copy(ala_sidechain)),
                                       (),
                                       (('peptide.CA',
                                         'sidechain.CB',
                                         'single'),))
        ala_dipeptide = M.Polymer('di_ala', 'alanine_dipeptide',
                                  (ala('ALA1'), ala('ALA2')),
                                  (('ALA1.peptide.C',
                                    'ALA2.peptide.N',
                                    'single'),),
                                  'polypeptide')
        self.universe = M.Universe('cube', [(ala_dipeptide, 2)],
                                   convention='my_own')
        self.assertTrue(is_valid(self.universe))


class IMPeptideTest(unittest.TestCase,
                    PeptideTest):
    M = mosaic.immutable_model

    def setUp(self):
        M = self.M
        C = M.element('C')
        H = M.element('H')
        N = M.element('N')
        O = M.element('O')
        peptide_group = M.fragment('peptide',
                                   (),
                                   (('CA', M.atom(C)),
                                    ('HA', M.atom(H)),
                                    ('H', M.atom(H)),
                                    ('N', M.atom(N)),
                                    ('C', M.atom(C)),
                                    ('O', M.atom(O))),
                                   (('N', 'H', "single"),
                                    ('N', 'CA', "single"),
                                    ('CA', 'HA', "single"),
                                    ('CA', 'C', "single"),
                                    ('C', 'O', "double")))
        ala_sidechain = M.fragment('ala_sidechain',
                                   (),
                                   (('CB', M.atom(C)),
                                    ('HB1', M.atom(H)),
                                    ('HB2', M.atom(H)),
                                    ('HB3', M.atom(H))),
                                   (('CB', 'HB1', "single"),
                                    ('CB', 'HB2', "single"),
                                    ('CB', 'HB3', "single"),))
        ala = M.fragment('alanine',
                         (('peptide', peptide_group),
                          ('sidechain', ala_sidechain)),
                         (),
                         (('peptide.CA', 'sidechain.CB', "single"),))
        ala_dipeptide = M.polymer('alanine_dipeptide',
                                  (('ALA1', ala),
                                   ('ALA2', ala)),
                                  (('ALA1.peptide.C', 'ALA2.peptide.N', "single"),),
                                  'polypeptide')
        self.universe = M.universe('cube', ((ala_dipeptide, 'di_ala', 2),),
                                   convention='my_own')
        self.assertTrue(is_valid(self.universe))

def suite():
    loader = unittest.TestLoader()
    s = unittest.TestSuite()
    s.addTest(loader.loadTestsFromTestCase(MMPeptideTest))
    s.addTest(loader.loadTestsFromTestCase(IMPeptideTest))
    return s

if __name__ == '__main__':
    unittest.main()
